# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------


"""
spyder.plugins.editor.widgets.codeeditor
========================================

Main Editor widget based on QtGui.QPlainTextEdit.
"""

from .codeeditor import (
    CodeEditor,
    CodeEditorActions,
    CodeEditorMenus,
    CodeEditorContextMenuSections,
    DocstringContext,
)



# -----------------------------------------------------------------------------
# Copyright (c) 2019- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""Automatically generate function docstrings."""

from __future__ import annotations

# Standard library imports
import re
import textwrap
from dataclasses import dataclass

# Third party imports
from qtpy.QtCore import Qt
from qtpy.QtGui import QTextCursor

# Local imports
from spyder.api.widgets.menus import SpyderMenu


# Constants

_MAX_CHARACTERS_PER_LINE = 120

MAX_SIG_LINES = 1000
"""Maximum number of lines scanned for function signatures."""

MAX_SIG_CHARS = MAX_SIG_LINES * _MAX_CHARACTERS_PER_LINE
"""Maximum number of characters in a function signature."""

MAX_RETURN_TYPE_LINES = 100
"""Maximum number of lines for the return type annotation."""

MAX_RETURN_TYPE_CHARS = MAX_RETURN_TYPE_LINES * _MAX_CHARACTERS_PER_LINE
"""Maximum number of characters in the return type annotation."""

MIN_HEADING = 3
"""Minimum number of characters in a docstring heading name."""

MAX_HEADING = 39
"""Maximum number of characters in a docstring heading name."""


def is_start_of_function(text):
    """Return True if text is the beginning of the function definition."""
    if isinstance(text, str):
        function_prefix = ['def ', 'async def ']
        text = text.lstrip()

        for prefix in function_prefix:
            if text.startswith(prefix):
                return True

    return False


def get_indent(text):
    """Get the leading whitespace characters of a line of text.

    Originally adapted from https://stackoverflow.com/q/2268532
    """
    match = re.match(r'(\s*)', text)
    return match.group(1) if match else ''


def get_common_indent(text):
    """Get the common indentation characters of a block of text."""
    indent = None
    lines = text.split("\n")
    for line in lines:
        if not line:
            continue
        line_indent = get_indent(line)
        if indent is None or len(line_indent) < len(indent):
            indent = line_indent
    return indent


def is_in_scope_forward(text):
    """Check if the next empty line could be part of the definition."""
    text = text.replace(r"\"", "").replace(r"\'", "")
    scopes = ["'''", '"""', "'", '"']
    indices = [MAX_SIG_CHARS] * 4
    for i in range(len(scopes)):
        if scopes[i] in text:
            indices[i] = text.index(scopes[i])
    if min(indices) == MAX_SIG_CHARS:
        return (text.count(")") != text.count("(") or
                text.count("]") != text.count("[") or
                text.count("}") != text.count("{"))

    s = scopes[indices.index(min(indices))]
    p = indices[indices.index(min(indices))]
    ls = len(s)
    if s in text[p + ls:]:
        text = text[:p] + text[p + ls:][text[p + ls:].index(s) + ls:]
        return is_in_scope_forward(text)
    if ls == 3:
        text = text[:p]
        return (text.count(")") != text.count("(") or
                text.count("]") != text.count("[") or
                text.count("}") != text.count("{"))

    return False


def is_tuple_brackets(text):
    """Check if the return type is a tuple."""
    scopes = ["(", "[", "{"]
    complements = [")", "]", "}"]
    indices = [MAX_RETURN_TYPE_CHARS] * 4
    for i in range(len(scopes)):
        if scopes[i] in text:
            indices[i] = text.index(scopes[i])
    if min(indices) == MAX_RETURN_TYPE_CHARS:
        return "," in text
    s = complements[indices.index(min(indices))]
    p = indices[indices.index(min(indices))]
    if s in text[p + 1:]:
        text = text[:p] + text[p + 1:][text[p + 1:].index(s) + 1:]
        return is_tuple_brackets(text)
    return False


def is_tuple_strings(text):
    """Check if the return type is a string."""
    text = text.replace(r"\"", "").replace(r"\'", "")
    scopes = ["'''", '"""', "'", '"']
    indices = [MAX_RETURN_TYPE_CHARS] * 4
    for i in range(len(scopes)):
        if scopes[i] in text:
            indices[i] = text.index(scopes[i])
    if min(indices) == MAX_RETURN_TYPE_CHARS:
        return is_tuple_brackets(text)
    s = scopes[indices.index(min(indices))]
    p = indices[indices.index(min(indices))]
    ls = len(s)
    if s in text[p + ls:]:
        text = text[:p] + text[p + ls:][text[p + ls:].index(s) + ls:]
        return is_tuple_strings(text)

    return False


def is_in_scope_backward(text):
    """Check if the next empty line could be part of the definition."""
    return is_in_scope_forward(
        text.replace(r"\"", "").replace(r"\'", "")[::-1])


def remove_comments(text):
    """
    Remove code comments from text.

    Ignores hash symbols (``#``) inside quotes, which can be part of
    function arguments.
    """
    return re.sub(pattern=r"""(?<!['"])(#.*)""", repl="", string=text)


def collapse_line_breaks(text):
    """Collapse line breaks into a single line."""
    text = text.strip(" \\").rstrip(",").replace(",]", "]").replace(",)", ")")
    lines = re.sub(r"\s{2,}", r"\n", text.strip()).split("\n")
    collapsed_lines = []
    for line in lines:
        line = line.strip()
        if not line:
            continue

        if not collapsed_lines:
            collapsed_lines.append(line)
            continue

        prev = collapsed_lines[-1][-1]
        this = line[0]
        if this in {")", "]", "}"}:
            collapsed_lines[-1] = collapsed_lines[-1].rstrip(",")
            collapsed_lines.append(line)
        elif prev in {"(", "[", "{"}:
            collapsed_lines.append(line)
        else:
            collapsed_lines.append(f" {line}")

    return "".join(collapsed_lines)


def collapse_line_breaks_annotation(text):
    """Collapse a type annotation into a single line."""
    text = text.strip().strip(""" "'()\\""").rstrip(',').replace(",]", "]")
    collapsed_lines = collapse_line_breaks(text)
    return collapsed_lines


def dedent_docstring(text):
    """Remove indentation up to the current scope."""
    if "\n" not in text:
        return text, None
    lines = text.split("\n")
    text_dedented = textwrap.dedent("\n".join(lines[1:]))
    text = "\n".join([lines[0], text_dedented])
    indent = get_common_indent("\n".join(lines[1:]))
    return text, indent


def indent_docstring(text, indent):
    """Add the specified indentation to all lines but the first."""
    if "\n" not in text:
        return text
    lines = text.split("\n")
    indent_lines = [indent + line if line else line for line in lines[1:]]
    text = "\n".join([lines[0], *indent_lines])
    return text


class DocstringWriterExtension:
    """Class for insert docstring template automatically."""

    def __init__(self, code_editor):
        """Initialize and Add code_editor to the variable."""
        self.code_editor = code_editor
        self.quote3 = '"""'
        self.quote3_other = "'''"
        self.line_number_cursor = None

    @staticmethod
    def is_beginning_triple_quotes(text):
        """Return True if there are only triple quotes in text."""
        docstring_triggers = ['"""', 'r"""', "'''", "r'''"]
        if text.lstrip() in docstring_triggers:
            return True

        return False

    def is_end_of_function_definition(self, text, line_number):
        """Return True if text is the end of the function definition."""
        text_without_whitespace = "".join(text.split())
        if (
            text_without_whitespace.endswith("):") or
            text_without_whitespace.endswith("]:") or
            (text_without_whitespace.endswith(":") and
             "->" in text_without_whitespace)
        ):
            return True

        if text_without_whitespace.endswith(":") and line_number > 1:
            complete_text = text_without_whitespace
            document = self.code_editor.document()
            cursor = QTextCursor(
                document.findBlockByNumber(line_number - 2))  # previous line

            for i in range(line_number - 2, -1, -1):
                txt = "".join(str(cursor.block().text()).split())
                if txt.endswith("\\") or is_in_scope_backward(complete_text):
                    if txt.endswith("\\"):
                        txt = txt[:-1]
                    complete_text = txt + complete_text
                else:
                    break
                if i:
                    cursor.movePosition(QTextCursor.PreviousBlock)

            if is_start_of_function(complete_text):
                return (
                    complete_text.endswith("):") or
                    complete_text.endswith("]:") or
                    (complete_text.endswith(":") and
                     "->" in complete_text)
                )
            return False

        return False

    def get_function_definition_from_first_line(self):
        """Get func def when the cursor is located on the first def line."""
        document = self.code_editor.document()
        cursor = QTextCursor(
            document.findBlockByNumber(self.line_number_cursor - 1))

        func_text = ''
        func_indent = ''

        is_first_line = True
        line_number = cursor.blockNumber() + 1

        number_of_lines = self.code_editor.blockCount()
        remain_lines = number_of_lines - line_number + 1
        number_of_lines_of_function = 0

        for __ in range(min(remain_lines, MAX_SIG_LINES)):
            cur_text = str(cursor.block().text()).rstrip()
            cur_text = remove_comments(cur_text)
            strip_text = cur_text.strip()

            if is_first_line:
                if not is_start_of_function(cur_text):
                    return None

                func_indent = get_indent(cur_text)
                is_first_line = False
            elif not strip_text:  # Skip empty lines
                pass
            else:
                cur_indent = get_indent(cur_text)
                if cur_indent < func_indent:  # Outside the function scope
                    return None
                if cur_indent == func_indent and not re.search(
                    r'^([\])]\s*:|\)\s*->.*[\[(:])$', strip_text
                ):  # Black-style function last line
                    return None
                if is_start_of_function(cur_text):
                    return None
                if not is_in_scope_forward(func_text):
                    return None

            if len(cur_text) > 0 and cur_text[-1] == '\\':
                cur_text = cur_text[:-1]

            func_text += cur_text
            number_of_lines_of_function += 1

            if self.is_end_of_function_definition(
                    cur_text, line_number + number_of_lines_of_function - 1):
                return func_text, number_of_lines_of_function

            cursor.movePosition(QTextCursor.NextBlock)

        return None

    def get_function_definition_from_below_last_line(self):
        """Get func def when the cursor is located below the last def line."""
        cursor = self.code_editor.textCursor()
        func_text = ''
        is_first_line = True
        line_number = cursor.blockNumber() + 1
        number_of_lines_of_function = 0

        for __ in range(min(line_number, MAX_SIG_LINES)):
            if cursor.block().blockNumber() == 0:
                return None

            cursor.movePosition(QTextCursor.PreviousBlock)
            prev_text = str(cursor.block().text()).rstrip()
            prev_text = remove_comments(prev_text)

            if is_first_line:
                if not self.is_end_of_function_definition(
                        prev_text, line_number - 1):
                    return None
                is_first_line = False
            elif self.is_end_of_function_definition(
                    prev_text, line_number - number_of_lines_of_function - 1):
                return None

            if len(prev_text) > 0 and prev_text[-1] == '\\':
                prev_text = prev_text[:-1]

            func_text = prev_text + func_text

            number_of_lines_of_function += 1
            if is_start_of_function(prev_text):
                return func_text, number_of_lines_of_function

        return None

    def get_function_docstring(self, func_indent, delete_existing=True):
        """Get the function's existing docstring content."""
        cursor = self.code_editor.textCursor()
        line_number = cursor.blockNumber()
        number_of_lines = self.code_editor.blockCount()
        docstring_list = []
        docstring_quotes = None
        last_line = False

        cursor.clearSelection()
        for __ in range(number_of_lines - line_number):
            cursor.movePosition(QTextCursor.NextBlock, QTextCursor.KeepAnchor)
            if last_line:
                break

            text = str(cursor.block().text())
            text_indent = get_indent(text)

            # Stop if outside the function scope
            if text.strip() and len(text_indent) <= len(func_indent):
                break

            # Look for the start of the docstring
            if not docstring_quotes:
                if not remove_comments(text).strip():
                    continue
                elif '"""' in text:
                    docstring_quotes = '"""'
                elif "'''" in text:
                    docstring_quotes = "'''"
                else:  # Found function body, no docstring
                    return None

                # One line docstring
                if text.count(docstring_quotes) > 1:
                    last_line = True
                cursor.clearSelection()
            # Look for the end
            elif docstring_quotes in text:
                last_line = True

            docstring_list.append(text)

            # So the final line is selected if docstring ends at the EOF
            cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.KeepAnchor)

        # If docstring was found and terminated
        if last_line:
            if delete_existing:
                cursor.removeSelectedText()
            cursor.clearSelection()
            return '\n'.join(docstring_list)
        else:
            cursor.clearSelection()
            return None

    def get_function_body(self, func_indent):
        """Get the function body text."""
        cursor = self.code_editor.textCursor()
        line_number = cursor.blockNumber()
        number_of_lines = self.code_editor.blockCount()
        body_list = []

        for __ in range(number_of_lines - line_number):
            text = str(cursor.block().text())
            text_indent = get_indent(text)

            # Stop if outside the function scope
            if text.strip() and len(text_indent) <= len(func_indent):
                break

            body_list.append(text)

            cursor.movePosition(QTextCursor.NextBlock)

        return '\n'.join(body_list)

    def write_docstring(self):
        """Write docstring to editor."""
        line_to_cursor = self.code_editor.get_text('sol', 'cursor')
        if not self.is_beginning_triple_quotes(line_to_cursor):
            return False

        cursor = self.code_editor.textCursor()
        prev_pos = cursor.position()
        quote = line_to_cursor[-1]
        docstring_type = self.code_editor.get_conf('docstring_type')

        cursor.beginEditBlock()
        try:
            docstring = self._generate_docstring(docstring_type, quote)
        except Exception:
            cursor.endEditBlock()
            raise

        if not docstring:
            cursor.endEditBlock()
            return False

        self.code_editor.insert_text(docstring)
        cursor.endEditBlock()

        # Set cursor to first line of summary
        cursor.setPosition(prev_pos, QTextCursor.KeepAnchor)
        cursor.movePosition(QTextCursor.NextBlock)
        cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
        cursor.clearSelection()
        self.code_editor.setTextCursor(cursor)

        return True

    def write_docstring_at_first_line_of_function(self):
        """Write docstring to editor at mouse position."""
        func_def_info = self.get_function_definition_from_first_line()
        editor = self.code_editor
        cursor = editor.textCursor()

        if func_def_info:
            func_text, number_of_line_func = func_def_info
            func_last_line_number = (
                self.line_number_cursor + number_of_line_func - 1
            )

            cursor_line_number = cursor.blockNumber() + 1
            offset = func_last_line_number - cursor_line_number
            if offset > 0:
                for __ in range(offset):
                    cursor.movePosition(QTextCursor.NextBlock)
            else:
                for __ in range(abs(offset)):
                    cursor.movePosition(QTextCursor.PreviousBlock)
            cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.MoveAnchor)
            editor.setTextCursor(cursor)

            body_indent = get_indent(func_text) + editor.indent_chars

            cursor.beginEditBlock()
            editor.insert_text(f'\n{body_indent}"""')
            try:
                self.write_docstring()
            except Exception:
                cursor.endEditBlock()
                editor.undo()
                raise
            else:
                cursor.endEditBlock()

    def write_docstring_for_shortcut(self):
        """Write docstring to editor by shortcut of code editor."""
        func_def_info = self.get_function_definition_from_below_last_line()
        cursor = self.code_editor.textCursor()

        # Cursor placed below function definition
        if func_def_info is not None:
            __, number_of_lines_of_function = func_def_info
            for __ in range(number_of_lines_of_function):
                cursor.movePosition(QTextCursor.PreviousBlock)

            self.code_editor.setTextCursor(cursor)

        self.line_number_cursor = cursor.blockNumber() + 1

        self.write_docstring_at_first_line_of_function()

    def _generate_docstring(self, doc_type, quote):
        """Generate a function/method docstring of the specified format."""
        docstring = None

        self.quote3 = quote * 3
        if quote == '"':
            self.quote3_other = "'''"
        else:
            self.quote3_other = '"""'

        func_def_info = self.get_function_definition_from_below_last_line()

        if not func_def_info:
            return None

        func_def, __ = func_def_info
        func_info = FunctionInfo()
        func_info.parse_def(func_def)

        if not func_info.has_info:
            return None

        func_docstring = self.get_function_docstring(
            func_info.func_indent
        )
        doc_info = DocstringInfo()
        doc_info.parse_docstring(func_docstring)

        func_body = self.get_function_body(func_info.func_indent)
        if func_body:
            func_info.parse_body(func_body)

        doc_type = doc_info.format_name or doc_type
        if doc_type == "Numpydoc":
            docstring = self._generate_numpy_doc(func_info, doc_info)
        elif doc_type == "Googledoc":
            docstring = self._generate_google_doc(func_info, doc_info)
        elif doc_type == "Sphinxdoc":
            docstring = self._generate_sphinx_doc(func_info, doc_info)
        else:
            raise ValueError(f"Unknown docstring format {doc_type!r}")

        return docstring

    def _generate_numpy_doc(self, func_info, doc_info):
        """Generate a NumPy format docstring."""
        numpy_doc = ''

        indent1 = func_info.func_indent + self.code_editor.indent_chars
        indent2 = self.code_editor.indent_chars

        summary = doc_info.description or 'SUMMARY.'
        numpy_doc += f'\n{indent1}{summary}\n'

        for section_fn, existing_section in [
            (self._generate_numpy_param_section, doc_info.parameters),
            (self._generate_numpy_return_section, doc_info.returns),
            (self._generate_numpy_raise_section, doc_info.raises),
            (lambda *__: None, doc_info.other),
        ]:
            doc_section = (
                existing_section or section_fn(func_info, indent1, indent2)
            )
            if doc_section:
                numpy_doc += f'\n{indent1}{doc_section}\n'

        numpy_doc = numpy_doc.rstrip()
        numpy_doc += f'\n{indent1}{self.quote3}'

        return numpy_doc

    def _generate_numpy_param_section(self, func_info, indent1, indent2):
        """Generate the Parameters section for a NumPy format docstring."""
        arg_names = func_info.arg_name_list.copy()
        arg_types = func_info.arg_type_list.copy()
        arg_values = func_info.arg_value_list.copy()

        if arg_names and arg_names[0] in ('self', 'cls'):
            del arg_names[0]
            del arg_types[0]
            del arg_values[0]

        if not arg_names:
            return None

        heading = 'Parameters'
        header_lines = [heading, '-' * len(heading)]

        body_lines = []
        for arg_name, arg_type, arg_value in zip(
            arg_names, arg_types, arg_values
        ):
            param_type = arg_type or 'TYPE'
            optional = ', optional' if arg_value else ''
            type_line = f'{arg_name} : {param_type}{optional}'

            default = ''
            if arg_value:
                arg_value = arg_value.replace(self.quote3, self.quote3_other)
                default = f' The default is {arg_value}.'
            desc_line = f'{indent2}DESCRIPTION.{default}'

            body_lines += [type_line, desc_line]

        return f'\n{indent1}'.join(header_lines + body_lines)

    def _generate_numpy_return_section(self, func_info, indent1, indent2):
        """Generate the Returns section for a NumPy format docstring."""
        heading = 'Yields' if func_info.has_yield else 'Returns'
        header = f'{heading}\n{indent1}{"-" * len(heading)}\n'
        indent3 = indent1 + indent2

        return_types = func_info.return_type_annotated
        if return_types:
            if len(return_types) > 1:
                return_type_annotated = (
                    f'\n{indent3}DESCRIPTION.\n{indent1}'.join(return_types)
                )
            else:
                return_type_annotated = return_types[0]
            return_section = f'{header}{indent1}{return_type_annotated}'
            if return_type_annotated != 'None':
                return_section += f'\n{indent3}DESCRIPTION.'
            return return_section

        return_element_type = (
            indent1 + '{return_type}\n' + indent3 + 'DESCRIPTION.'
        )
        placeholder = return_element_type.format(return_type='TYPE')
        return_values = [
            rv for rvs in func_info.return_value_in_body
            for rv in rvs.split(",")
        ]
        if len(return_values) == 1:
            # numpydoc RT02, only include type and not name if it's a single return value
            return_element_name = indent1 + placeholder.lstrip()
        else:
            return_element_name = (
                indent1 + '{return_name} : ' + placeholder.lstrip()
            )

        try:
            return_section = self._generate_docstring_return_section(
                return_vals=func_info.return_value_in_body,
                header=header,
                return_element_name=return_element_name,
                return_element_type=return_element_type,
                placeholder=placeholder,
                indent=indent1,
                expand_tuple=True,
            )
        except (ValueError, IndexError):
            return_section = f'{header}{indent1}None'

        return return_section

    @staticmethod
    def _generate_numpy_raise_section(func_info, indent1, indent2):
        """Generate the Raises section for a NumPy format docstring."""
        if not func_info.raise_list:
            return None

        heading = 'Raises'
        header_lines = [heading, '-' * len(heading)]
        body_lines = [
            [f'{raise_type}', f'{indent2}DESCRIPTION.']
            for raise_type in func_info.raise_list
        ]

        return f'\n{indent1}'.join(sum(body_lines, header_lines))

    def _generate_google_doc(self, func_info, doc_info):
        """Generate a Google format docstring."""
        google_doc = ''

        indent1 = func_info.func_indent + self.code_editor.indent_chars
        indent2 = self.code_editor.indent_chars

        summary = doc_info.description or 'SUMMARY.'
        google_doc += f'{summary}\n'

        for section_fn, existing_section in [
            (self._generate_google_param_section, doc_info.parameters),
            (self._generate_google_return_section, doc_info.returns),
            (self._generate_google_raise_section, doc_info.raises),
            (lambda *__: None, doc_info.other),
        ]:
            doc_section = (
                existing_section or section_fn(func_info, indent1, indent2)
            )
            if doc_section:
                google_doc += f'\n{indent1}{doc_section}\n'


        google_doc = google_doc.rstrip()
        google_doc += f'\n{indent1}{self.quote3}'

        return google_doc

    def _generate_google_param_section(self, func_info, indent1, indent2):
        """Generate the Args section for a Google format docstring."""
        arg_names = func_info.arg_name_list.copy()
        arg_types = func_info.arg_type_list.copy()
        arg_values = func_info.arg_value_list.copy()

        if arg_names and arg_names[0] in ('self', 'cls'):
            del arg_names[0]
            del arg_types[0]
            del arg_values[0]

        if not arg_names:
            return None

        header_lines = ['Args:']

        body_lines = []
        for arg_name, arg_type, arg_value in zip(
            arg_names, arg_types, arg_values
        ):
            param_type = arg_type or 'TYPE'
            optional = ', optional' if arg_value else ''
            type_chunk = f'{indent2}{arg_name} ({param_type}{optional})'

            default = ''
            if arg_value:
                arg_value = arg_value.replace(self.quote3, self.quote3_other)
                default = f' Defaults to {arg_value}.'
            desc_chunk = f'DESCRIPTION.{default}'

            param_line = f'{type_chunk}: {desc_chunk}'
            body_lines.append(param_line)

        return f'\n{indent1}'.join(header_lines + body_lines)

    def _generate_google_return_section(self, func_info, indent1, indent2):
        """Generate the Returns section for a Google format docstring."""
        heading = 'Yields' if func_info.has_yield else 'Returns'
        header = f'{heading}:\n'
        indent3 = indent1 + indent2

        return_types = func_info.return_type_annotated
        if return_types:
            if len(return_types) > 1:
                tuple_values = ', '.join(return_types)
                return_type_annotated = f'tuple[{tuple_values}]'
            else:
                return_type_annotated = return_types[0]

            return_section = f'{header}{indent3}{return_type_annotated}'
            if return_type_annotated != 'None':
                return_section += ': DESCRIPTION.'

            return return_section

        return_element_type = indent3 + '{return_type}: DESCRIPTION.'
        placeholder = return_element_type.format(return_type='TYPE')

        try:
            return_section = self._generate_docstring_return_section(
                return_vals=func_info.return_value_in_body,
                header=header,
                return_element_name=placeholder,
                return_element_type=return_element_type,
                placeholder=placeholder,
                indent=indent3,
                expand_tuple=False,
            )
        except (ValueError, IndexError):
            return_section = f'{header}{indent3}None'

        return return_section

    @staticmethod
    def _generate_google_raise_section(func_info, indent1, indent2):
        """Generate the Raises section for a Google format docstring."""
        if not func_info.raise_list:
            return None

        header_lines = ['Raises:']
        body_lines = [
            f'{indent2}{raise_type}: DESCRIPTION.'
            for raise_type in func_info.raise_list
        ]

        return f'\n{indent1}'.join(header_lines + body_lines)

    def _generate_sphinx_doc(self, func_info, doc_info):
        """Generate a Sphinx format docstring."""
        sphinx_doc = ''
        indent1 = func_info.func_indent + self.code_editor.indent_chars

        summary = doc_info.description or 'SUMMARY.'
        sphinx_doc += f'{summary}\n'

        for section_fn, existing_section in [
            (self._generate_sphinx_param_section, doc_info.parameters),
            (self._generate_sphinx_return_section, doc_info.returns),
            (self._generate_sphinx_raise_section, doc_info.raises),
            (lambda *__: None, doc_info.other),
        ]:
            doc_section = (
                existing_section or section_fn(func_info, indent1)
            )
            if doc_section:
                sphinx_doc += f'\n{indent1}{doc_section}\n'

        sphinx_doc = sphinx_doc.rstrip()
        sphinx_doc += f'\n{indent1}{self.quote3}'

        return sphinx_doc

    def _generate_sphinx_param_section(self, func_info, indent1):
        """Generate the :param: sections for a Sphinx format docstring."""
        arg_names = func_info.arg_name_list.copy()
        arg_types = func_info.arg_type_list.copy()
        arg_values = func_info.arg_value_list.copy()

        if arg_names and arg_names[0] in ('self', 'cls'):
            del arg_names[0]
            del arg_types[0]
            del arg_values[0]

        if not arg_names:
            return None

        param_lines = []
        for arg_name, arg_type, arg_value in zip(
            arg_names, arg_types, arg_values
        ):
            param_desc = f':param {arg_name}: DESCRIPTION'

            if arg_value:
                arg_value = arg_value.replace(self.quote3, self.quote3_other)
                param_desc += f', defaults to {arg_value}'

            param_type = f':type {arg_name}: '

            if arg_type:
                param_type += f'{arg_type}'
            else:
                param_type += 'TYPE'

            param_lines += [param_desc, param_type]

        return f'\n{indent1}'.join(param_lines)

    def _generate_sphinx_return_section(self, func_info, indent1):
        """Generate the :return: section for a Sphinx format docstring."""
        header = ':rtype: '
        return_desc = f'{indent1}:returns: DESCRIPTION'

        return_types = func_info.return_type_annotated
        if return_types:
            if len(return_types) > 1:
                tuple_values = ', '.join(return_types)
                return_type_annotated = f'tuple[{tuple_values}]'
            else:
                return_type_annotated = return_types[0]
            return_section = f'{header}{return_type_annotated}'
            if return_type_annotated != 'None':
                return_section += f'\n{return_desc}'
            return return_section

        return_element_type = f'{{return_type}}\n{return_desc}'
        placeholder = return_element_type.format(return_type='TYPE')

        try:
            return_section = self._generate_docstring_return_section(
                return_vals=func_info.return_value_in_body,
                header=header,
                return_element_name=placeholder,
                return_element_type=return_element_type,
                placeholder=placeholder,
                indent="",
                expand_tuple=False,
            )
        except (ValueError, IndexError):
            return_section = f'{header}None'

        return return_section

    @staticmethod
    def _generate_sphinx_raise_section(func_info, indent1):
        """Generate the :raises: sections for a Sphinx format docstring."""
        if not func_info.raise_list:
            return None

        raise_lines = [
            f':raises {raise_type}: DESCRIPTION'
            for raise_type in func_info.raise_list
        ]

        return f'\n{indent1}'.join(raise_lines)

    @staticmethod
    def find_top_level_bracket_locations(string_toparse):
        """Get the locations of top-level brackets in a string."""
        bracket_stack = []
        replace_args_list = []
        bracket_type = None
        literal_type = ''
        brackets = {'(': ')', '[': ']', '{': '}'}
        for idx, character in enumerate(string_toparse):
            if (not bracket_stack and character in brackets.keys()
                    or character == bracket_type):
                bracket_stack.append(idx)
                bracket_type = character
            elif bracket_type and character == brackets[bracket_type]:
                begin_idx = bracket_stack.pop()
                if not bracket_stack:
                    if not literal_type:
                        if bracket_type == '(':
                            literal_type = '(None)'
                        elif bracket_type == '[':
                            literal_type = '[list]'
                        elif bracket_type == '{':
                            if idx - begin_idx <= 1:
                                literal_type = '{dict}'
                            else:
                                literal_type = '{set}'
                    replace_args_list.append(
                        (string_toparse[begin_idx:idx + 1],
                         literal_type, 1))
                    bracket_type = None
                    literal_type = ''
            elif len(bracket_stack) == 1:
                if bracket_type == '(' and character == ',':
                    literal_type = '(tuple)'
                elif bracket_type == '{' and character == ':':
                    literal_type = '{dict}'
                elif bracket_type == '(' and character == ':':
                    literal_type = '[slice]'

        if bracket_stack:
            raise IndexError('Bracket mismatch')
        for replace_args in replace_args_list:
            string_toparse = string_toparse.replace(*replace_args)
        return string_toparse

    @staticmethod
    def parse_return_elements(return_vals_group):
        """Return the appropriate text for a group of return elements."""
        all_eq = return_vals_group.count(return_vals_group[0]) == len(
            return_vals_group
        )
        builtin_collections = {"[list]", "(tuple)", "{dict}", "{set}"}
        if builtin_collections.issuperset(return_vals_group) and all_eq:
            return return_vals_group[0][1:-1], None

        # Output placeholder if special Python chars present in name
        py_chars = {' ', '+', '-', '*', '/', '%', '@', '<', '>', '&', '|', '^',
                    '~', '=', ',', ':', ';', '#', '(', '[', '{', '}', ']',
                    ')', }
        if any(
            any(py_char in return_val for py_char in py_chars)
            for return_val in return_vals_group
        ):
            return None, None

        # Output str type and no name if only string literals
        if all(
            '"' in return_val or "'" in return_val
            for return_val in return_vals_group
        ):
            return 'str', None

        # Output bool type and no name if only bool literals
        if {'True', 'False'}.issuperset(return_vals_group):
            return 'bool', None

        # Output numeric types and no name if only numeric literals
        try:
            [float(return_val) for return_val in return_vals_group]
            num_not_int = 0
            for return_val in return_vals_group:
                try:
                    int(return_val)
                except ValueError:  # If not an integer (EAFP)
                    num_not_int += 1
            if num_not_int == 0:
                return 'int', None
            if num_not_int == len(return_vals_group):
                return 'float', None
            return 'numeric', None

        except ValueError:  # Not a numeric if float conversion didn't work
            pass

        # If names are not equal, don't contain "." or are a builtin
        if (
            {"self", "cls", "None"}.isdisjoint(return_vals_group)
            and all_eq
            and all("." not in return_val for return_val in return_vals_group)
        ):
            return None, return_vals_group[0]

        return None, None

    def _generate_docstring_return_section(
        self,
        return_vals,
        header,
        return_element_name,
        return_element_type,
        placeholder,
        indent,
        *,
        expand_tuple=True,
    ):
        """Generate the Returns section of a function/method docstring."""
        # If all return values are None, return None
        non_none_vals = [return_val for return_val in return_vals
                         if return_val and return_val != 'None']
        if not non_none_vals:
            return header + indent + 'None'

        # Get only values with matching brackets that can be cleaned up
        non_none_vals = [
            return_val.strip(' ()\t\n').rstrip(',')
            for return_val in non_none_vals
        ]
        non_none_vals = [
            re.sub(r'''(["'])(?:(?=(\\?))\2.)*?\1''', r'"string"', return_val)
            for return_val in non_none_vals
        ]
        unambiguous_vals = []
        for return_val in non_none_vals:
            try:
                cleaned_val = self.find_top_level_bracket_locations(return_val)
            except IndexError:
                continue
            unambiguous_vals.append(cleaned_val)
        if not unambiguous_vals:
            return header + placeholder

        # If remaining are a mix of tuples and not, return single placeholder
        single_vals, tuple_vals = [], []
        for return_val in unambiguous_vals:
            if ',' in return_val:
                tuple_vals.append(return_val)
            else:
                single_vals.append(return_val)

        if single_vals and tuple_vals:
            return header + placeholder

        # If return values are tuples of different length, return a placeholder
        if tuple_vals:
            num_elements = [return_val.count(',') + 1
                            for return_val in tuple_vals]
            if num_elements.count(num_elements[0]) != len(num_elements):
                return header + placeholder
            num_elements = num_elements[0]
        else:
            num_elements = 1

        # If all have the same len but some ambiguous return placeholders
        if len(unambiguous_vals) != len(non_none_vals):
            if expand_tuple:
                return header + '\n'.join([placeholder] * len(num_elements))

            return_elements_out = ', '.join(['TYPE'] * len(num_elements))
            return header + return_element_type.format(
                return_type=f'tuple[{return_elements_out}]'
            )

        # Handle tuple (or single) values position by position
        return_vals_grouped = zip(*[
            [return_element.strip() for return_element in
             return_val.split(',')]
            for return_val in unambiguous_vals])
        return_vals_parsed = []
        for return_vals_group in return_vals_grouped:
            return_vals_parsed.append(
                self.parse_return_elements(return_vals_group)
            )

        # Represent a tuple as a single tuple return value
        if not expand_tuple:
            return_elements = [
                rtype if rtype is not None else 'TYPE'
                for rtype, __ in return_vals_parsed
            ]
            if len(return_elements) == 1:
                return_type = return_elements[0]
            else:
                return_type = 'tuple[{}]'.format(', '.join(return_elements))
            return header + return_element_type.format(return_type=return_type)

        # Represent a tuple as multiple return values
        return_elements = []
        for rtype, rname in return_vals_parsed:
            if rtype is not None:
                element = return_element_type.format(return_type=rtype)
            elif rname is not None:
                element = return_element_name.format(return_name=rname)
            else:
                element = placeholder
            return_elements.append(element)

        return header + '\n'.join(return_elements)


class FunctionInfo:
    """Parse function definition text."""

    RETURN_TYPE_REGEX = (
        r'->\s*((?:\s|[\"\'a-zA-Z0-9_,.()\[\]|])+)\s*\: *$'
    )
    RETURN_TUPLE_REGEX = r'^(?:typing\.)?[Tt]uple\[\s*((?:\s|.)+)\s*\]$'

    def __init__(self):
        self.has_info = False
        self.args_text = ''
        self.func_indent = ''
        self.arg_name_list = []
        self.arg_type_list = []
        self.arg_value_list = []
        self.return_type_annotated = None
        self.return_value_in_body = []
        self.raise_list = None
        self.has_yield = False

    @staticmethod
    def is_char_in_pairs(pos_char, pairs):
        """Return True if the character is in pairs of brackets or quotes."""
        for pos_left, pos_right in pairs.items():
            if pos_left < pos_char < pos_right:
                return True

        return False

    @staticmethod
    def _find_quote_position(text):
        """Return the start and end position of pairs of quotes."""
        pos = {}
        is_found_left_quote = False
        quote = None
        left_pos = None

        for idx, character in enumerate(text):
            if not is_found_left_quote:
                if character in {"'", '"'}:
                    is_found_left_quote = True
                    quote = character
                    left_pos = idx
            else:
                # Find (unescaped) quotes
                if character == quote and (
                    text[idx - 1] != "\\" or text[idx - 2] == "\\"
                ):
                    pos[left_pos] = idx
                    is_found_left_quote = False

        if is_found_left_quote:
            raise IndexError(f"No matching close quote at: {left_pos}")

        return pos

    def _find_bracket_position(
        self, text, bracket_left, bracket_right, pos_quote
    ):
        """Return the start and end position of pairs of brackets.

        Originally adapted from https://stackoverflow.com/q/29991917
        """
        pos = {}
        pstack = []

        for idx, character in enumerate(text):
            if character == bracket_left and not self.is_char_in_pairs(
                idx, pos_quote
            ):
                pstack.append(idx)
            elif character == bracket_right and not self.is_char_in_pairs(
                idx, pos_quote
            ):
                if not len(pstack):
                    raise IndexError(f"No matching closing parens at: {idx}")
                pos[pstack.pop()] = idx

        if len(pstack) > 0:
            raise IndexError(f"No matching opening parens at: {pstack.pop()}")

        return pos

    def split_arg_to_name_type_value(self, args_list):
        """Split argument text to name, type, value."""
        for arg in args_list:
            # Skip arg if it is a `*` or `/` separator
            if arg.strip() in {'*', '/'}:
                continue

            arg_type = None
            arg_value = None

            has_type = False
            has_value = False

            pos_colon = arg.find(':')
            pos_equal = arg.find('=')

            if pos_equal > -1:
                has_value = True

            if pos_colon > -1:
                if not has_value:
                    has_type = True
                elif pos_equal > pos_colon:  # exception for def foo(arg1=":")
                    has_type = True

            if has_value and has_type:
                arg_name = arg[0:pos_colon].strip()
                arg_type = arg[pos_colon + 1:pos_equal].strip()
                arg_value = arg[pos_equal + 1:].strip()
            elif not has_value and has_type:
                arg_name = arg[0:pos_colon].strip()
                arg_type = arg[pos_colon + 1:].strip()
            elif has_value and not has_type:
                arg_name = arg[0:pos_equal].strip()
                arg_value = arg[pos_equal + 1:].strip()
            else:
                arg_name = arg.strip()

            if arg_type:
                arg_type = collapse_line_breaks_annotation(arg_type)
            if arg_value:
                arg_value = collapse_line_breaks(arg_value)

            self.arg_name_list.append(arg_name)
            self.arg_type_list.append(arg_type)
            self.arg_value_list.append(arg_value)

    def split_args_text_to_list(self, args_text):
        """Split the text including multiple arguments to list.

        This function uses a comma to separate arguments and ignores a comma in
        brackets and quotes.
        """
        args_list = []
        idx_find_start = 0
        idx_arg_start = 0

        try:
            pos_quote = self._find_quote_position(args_text)
            pos_round = self._find_bracket_position(
                args_text, '(', ')', pos_quote
            )
            pos_curly = self._find_bracket_position(
                args_text, '{', '}', pos_quote
            )
            pos_square = self._find_bracket_position(
                args_text, '[', ']', pos_quote
            )
        except IndexError:
            return None

        while True:
            pos_comma = args_text.find(',', idx_find_start)

            if pos_comma == -1:
                break

            idx_find_start = pos_comma + 1

            if (
                self.is_char_in_pairs(pos_comma, pos_round)
                or self.is_char_in_pairs(pos_comma, pos_curly)
                or self.is_char_in_pairs(pos_comma, pos_square)
                or self.is_char_in_pairs(pos_comma, pos_quote)
            ):
                continue

            args_list.append(args_text[idx_arg_start:pos_comma])
            idx_arg_start = pos_comma + 1

        if idx_arg_start < len(args_text):
            arg_text = args_text[idx_arg_start:]
            # Skip arg if its empty (e.g. trailing comma)
            if arg_text.strip():
                args_list.append(arg_text)

        return args_list

    def split_return_tuple(self, text):
        """Split the type variables to a return tuple."""
        return_items = self.split_args_text_to_list(text)
        return_items_stripped = []

        for item in return_items:
            item = item.strip().strip(""" "'""")
            if item and item != '...':
                return_items_stripped.append(item)

        return return_items_stripped

    def parse_return_type_annotation(self, text):
        """Extract, parse and format the function's return type annotation."""
        return_type_match = re.search(self.RETURN_TYPE_REGEX, text)

        if not return_type_match:
            return None, len(text)

        return_type = return_type_match.group(1).strip()
        return_type = collapse_line_breaks_annotation(return_type)
        text_end = text.rfind(return_type_match.group(0))

        # If not a tuple, return the whole type
        return_tuple_match = re.search(self.RETURN_TUPLE_REGEX, return_type)
        if not return_tuple_match:
            return [return_type], text_end

        # If a 1-tuple, return the whole type
        tuple_values = return_tuple_match.group(1).strip()
        if not is_tuple_brackets(tuple_values):
            return [return_type], text_end

        # If only one item or an arbitrary-length tuple, return the whole type
        return_types = self.split_return_tuple(tuple_values)
        if len(return_types) < 2:
            return [return_type], text_end

        # Else, return the list of return tuple items
        return return_types, text_end

    def parse_def(self, text):
        """Parse the function definition text."""
        self.__init__()

        if not is_start_of_function(text):
            return

        self.func_indent = get_indent(text)

        text = text.strip()
        self.return_type_annotated, text_end = (
            self.parse_return_type_annotation(text)
        )

        pos_args_start = text.find('(') + 1
        pos_args_end = text.rfind(')', pos_args_start, text_end)

        self.args_text = text[pos_args_start:pos_args_end]

        args_list = self.split_args_text_to_list(self.args_text)
        if args_list is not None:
            self.has_info = True
            self.split_arg_to_name_type_value(args_list)

    def parse_body(self, text):
        """Parse the function body text."""
        re_raise = re.findall(r'(?:^|\n)[ \t]+raise +(\w+)', text)
        if len(re_raise) > 0:
            raise_list = [x.strip() for x in re_raise]
            # Remove duplicates from list while keeping it in order
            self.raise_list = list(dict.fromkeys(raise_list))

        re_yield = re.search(r'(?:^|\n)[ \t]+yield +\S', text)
        if re_yield:
            self.has_yield = True

        # Extract return value
        pattern_return = r'yield +' if re_yield else r'return +'
        line_list = text.split('\n')
        is_found_return = False
        line_return_tmp = ''

        for line in line_list:
            line = line.strip()

            if is_found_return is False:
                if re.match(pattern_return, line):
                    is_found_return = True

            if is_found_return:
                line_return_tmp += line
                # check the integrity of line
                try:
                    pos_quote = self._find_quote_position(line_return_tmp)

                    if line_return_tmp[-1] == '\\':
                        line_return_tmp = line_return_tmp[:-1]
                        continue

                    self._find_bracket_position(
                        line_return_tmp, '(', ')', pos_quote
                    )
                    self._find_bracket_position(
                        line_return_tmp, "{", "}", pos_quote
                    )
                    self._find_bracket_position(
                        line_return_tmp, "[", "]", pos_quote
                    )
                except IndexError:
                    continue

                return_value = re.sub(pattern_return, '', line_return_tmp)
                self.return_value_in_body.append(return_value)

                is_found_return = False
                line_return_tmp = ''


@dataclass
class DocstringFormat:
    """Collection of static data for each known docstring format."""
    name: str
    param_sections: list[str]
    return_sections: list[str]
    raise_sections: list[str]
    other_sections: list[str] | None
    before_section_regex: str | None = None
    section_regex: str = r"^({})$"
    after_section_regex: str | None = None
    non_section_regex: str = None
    any_section_name_regex: str = fr"[A-Za-z ]{{{MIN_HEADING},{MAX_HEADING}}}"

    @staticmethod
    def _strip_multiline(pattern):
        """Convert a multi-line mode regex to a standard regex."""
        return pattern.lstrip(r"^").replace(r"$", r"\n")

    def _build_section_regex(self, sections: list[str]) -> str:
        """Build a regex for different section headings from the template."""
        return self.section_regex.format(r"|".join(sections))

    @property
    def param_section_regex(self) -> str:
        """Regex matching the Parameters section heading."""
        return self._build_section_regex(self.param_sections)

    @property
    def return_section_regex(self) -> str:
        """Regex matching the Returns section heading."""
        return self._build_section_regex(self.return_sections)

    @property
    def raise_section_regex(self) -> str:
        """Regex matching the Raises section heading."""
        return self._build_section_regex(self.raise_sections)

    @property
    def other_section_regex(self) -> str:
        """Regex matching other section headings."""
        if self.non_section_regex is None and self.other_sections is not None:
            return self._build_section_regex(self.other_sections)
        return self.non_section_regex

    @property
    def known_sections(self) -> list[str]:
        """List of section names specified by this docstring format."""
        return (
            self.param_sections
            + self.return_sections
            + self.raise_sections
            + (self.other_sections or [])
        )

    @property
    def known_section_regex(self) -> str:
        """Regex matching section names specified by this docstring format."""
        return self._build_section_regex(self.known_sections)

    @property
    def any_section_regex(self) -> str:
        """Regex matching any possible section of a docstring format."""
        if self.other_sections is None:
            return self.known_section_regex
        return self.section_regex.format(self.any_section_name_regex)

    @property
    def any_other_section_regex(self) -> str:
        """Regex matching any non-param/return/raise content."""
        if self.non_section_regex is not None:
            return self.non_section_regex
        return self.any_section_regex

    @property
    def detection_regex(self) -> str:
        """Regex that should match any docstring of the specified format."""
        regexes = [
            r"\n",
            self.before_section_regex or "",
            self.known_section_regex,
            self.after_section_regex or "",
        ]
        return "".join(self._strip_multiline(regex) for regex in regexes)


DOC_FORMATS = {
    "Numpydoc": DocstringFormat(
        name="Numpydoc",
        param_sections=["Parameters"],
        return_sections=["Returns", "Yields"],
        raise_sections=["Raises"],
        other_sections=[
            "Warns",
            "Warnings",
            "See Also",
            "Notes",
            "References",
            "Examples",
        ],
        before_section_regex=r"^$",
        section_regex=r"^({})$",
        after_section_regex=fr"^-{{{MIN_HEADING},{MAX_HEADING}}}$",
    ),
    "Googledoc": DocstringFormat(
        name="Googledoc",
        param_sections=["Args"],
        return_sections=["Returns", "Yields"],
        raise_sections=["Raises"],
        other_sections=["Example", "Examples"],
        before_section_regex=r"^$",
        section_regex=r"^({}):$",
        after_section_regex=r"^(\t|(  ){1,4})[A-Za-z]",
    ),
    "Sphinxdoc": DocstringFormat(
        name="Sphinxdoc",
        param_sections=[
            "param",
            "parameters",
            "arg",
            "arguments",
            "key",
            "keyword",
            "type",
        ],
        return_sections=["return", "returns", "rtype"],
        raise_sections=["raise", "raises", "except", "exception"],
        other_sections=None,
        before_section_regex=None,
        section_regex=r"^:({}):? ",
        after_section_regex=None,
        non_section_regex=r"^\S",
    ),
}


class DocstringInfo:
    """Parse docstring sections and content."""

    def __init__(self):
        self.text = None
        self.doc_indent = None
        self.doc_format = None
        self.format_name = None
        self.description = None
        self.parameters = None
        self.returns = None
        self.raises = None
        self.other = None

    @staticmethod
    def _strip_docstring(text):
        """Strip leading/trailing whitespace and quotes."""
        text = text.strip()
        for quotes in ['"""', "'''"]:
            text = text.removeprefix(quotes).removesuffix(quotes)
        text = text.strip()
        text = "\n".join(line.rstrip() for line in text.split("\n"))
        return text

    @staticmethod
    def _detect_docstring_format(text):
        """Detect the format (Sphinx, NumPy, Google) of a docstring."""
        # Reverse the format order to reduce the false positive chance
        for name, doc_format in reversed(DOC_FORMATS.items()):
            if re.search(doc_format.detection_regex, text):
                return name, doc_format
        return None, None

    def _lines_to_text(self, lines):
        """Convert a list of text lines to a block of docstring text."""
        text = "\n".join(lines).rstrip()
        return indent_docstring(text, self.doc_indent)

    def _match_each_line(self, before, line, after, line_regex):
        """Match before, line and after with its corresponding regex."""
        match_pairs = [
            (before, self.doc_format.before_section_regex),
            (line, line_regex),
            (after, self.doc_format.after_section_regex),
        ]
        for text, regex in match_pairs:
            if regex is not None and not re.search(regex, text):
                return False
        return True

    def _is_section_start(self, before, line, after):
        """Check if this line starts a new docstring section."""
        return self._match_each_line(
            before, line, after, self.doc_format.any_section_regex
        )

    def _parse_generic_doc(self, text):
        """Parse a generic docstring without a known format."""
        self.description = indent_docstring(text, self.doc_indent)

    def _parse_known_doc(self, text):
        """Parse a known format docstring into its sections."""
        desc_ended = False
        current_section = None

        description = []
        parameters = []
        returns = []
        raises = []
        other = []

        # Add dummy first and last lines to avoid special casing
        text_lines = ["", *text.split("\n"), ""]

        sections = [
            (self.doc_format.param_section_regex, parameters),
            (self.doc_format.return_section_regex, returns),
            (self.doc_format.raise_section_regex, raises),
            (self.doc_format.any_other_section_regex, other),
        ]

        for idx, line in enumerate(text_lines):
            # Skip dummy first and last lines
            if not idx or idx == len(text_lines) - 1:
                continue

            before = text_lines[idx - 1]
            after = text_lines[idx + 1]

            desc_ended = desc_ended or self._is_section_start(
                before, line, after
            )
            if not desc_ended:
                description.append(line)
                continue

            for regex, section_lines in sections:
                matched = self._match_each_line(before, line, after, regex)
                if matched:
                    current_section = section_lines
                    break

            current_section.append(line)

        self.description = self._lines_to_text(description)
        self.parameters = self._lines_to_text(parameters)
        self.returns = self._lines_to_text(returns)
        self.raises = self._lines_to_text(raises)
        self.other = self._lines_to_text(other)

    def parse_docstring(self, text):
        """Process a callable's docstring into a structured form."""
        if not text:
            self.text = False
            return

        text = self._strip_docstring(text)
        self.text = text

        # Detect existing docstring format
        dedented_text, self.doc_indent = dedent_docstring(text)
        self.format_name, self.doc_format = self._detect_docstring_format(
            dedented_text
        )

        if self.doc_format:
            self._parse_known_doc(dedented_text)
        else:
            self._parse_generic_doc(dedented_text)


class QMenuOnlyForEnter(SpyderMenu):
    """The class executes the selected action when "enter key" is input.

    If a input of keyboard is not the "enter key", the menu is closed and
    the input is inserted to code editor.
    """

    def __init__(self, code_editor):
        """Init SpyderMenu."""
        super().__init__(code_editor)
        self.code_editor = code_editor

    def keyPressEvent(self, event):
        """Close the instance if key is not enter key."""
        key = event.key()
        if key not in (Qt.Key_Enter, Qt.Key_Return):
            self.code_editor.keyPressEvent(event)
            self.close()
        else:
            super().keyPressEvent(event)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2019- Spyder Project Contributors
#
# Components of objectbrowser originally distributed under
# the MIT (Expat) license. Licensed under the terms of the MIT License;
# see NOTICE.txt in the Spyder root directory for details
# -----------------------------------------------------------------------------

# Standard library imports
import copy
import datetime
import functools
import math
import operator
import sys
import traceback
from typing import Any, Callable, Optional

# Third party imports
from qtpy.compat import to_qvariant
from qtpy.QtCore import (
    QDateTime,
    QEvent,
    QItemSelection,
    QItemSelectionModel,
    QModelIndex,
    QRect,
    QSize,
    Qt,
    Signal,
)
from qtpy.QtWidgets import (
    QAbstractItemDelegate,
    QApplication,
    QDateEdit,
    QDateTimeEdit,
    QItemDelegate,
    QLineEdit,
    QMessageBox,
    QStyle,
    QStyleOptionButton,
    QTableView,
)
from spyder_kernels.comms.commbase import CommsErrorWrapper
from spyder_kernels.utils.lazymodules import (
    FakeObject, numpy as np, pandas as pd, PIL)
from spyder_kernels.utils.nsview import (display_to_value, is_editable_type,
                                         is_known_type)

# Local imports
from spyder.api.fonts import SpyderFontsMixin, SpyderFontType
from spyder.api.translations import _
from spyder.plugins.variableexplorer.widgets.arrayeditor import ArrayEditor
from spyder.plugins.variableexplorer.widgets.dataframeeditor import (
    DataFrameEditor)
from spyder.plugins.variableexplorer.widgets.texteditor import TextEditor
from spyder.utils.icon_manager import ima


LARGE_COLLECTION = 1e5
LARGE_ARRAY = 5e6
SELECT_ROW_BUTTON_SIZE = 22

# The maximum length of the text in a QLineEdit is 32_767, so do not allow
# the user to edit an `int` with more than 32_767 digits
MAX_INT_DIGITS_FOR_EDITING = 32_767
MAX_INT_BIT_LENGTH_FOR_EDITING = (
    int(MAX_INT_DIGITS_FOR_EDITING * math.log(10) / math.log(2))
)


class CollectionsDelegate(QItemDelegate, SpyderFontsMixin):
    """CollectionsEditor Item Delegate"""
    sig_free_memory_requested = Signal()
    sig_editor_creation_started = Signal()
    sig_editor_shown = Signal()

    def __init__(
        self,
        parent=None,
        namespacebrowser=None,
        data_function: Optional[Callable[[], Any]] = None
    ):
        QItemDelegate.__init__(self, parent)
        self.namespacebrowser = namespacebrowser
        self.data_function = data_function
        self._editors = {}  # keep references on opened editors

        try:
            if sys.get_int_max_str_digits() < MAX_INT_DIGITS_FOR_EDITING:
                sys.set_int_max_str_digits(MAX_INT_DIGITS_FOR_EDITING)
        except AttributeError:
            # There was no limit in Python 3.10 and earlier
            pass

    def get_value(self, index):
        if index.isValid():
            return index.model().get_value(index)

    def set_value(self, index, value):
        if index.isValid():
            index.model().set_value(index, value)

    def make_data_function(
        self,
        index: QModelIndex
    ) -> Optional[Callable[[], Any]]:
        """
        Construct function which returns current value of data.

        This is used to refresh editors created from this piece of data.
        For instance, if `self` is the delegate for an editor that displays
        the dict `xxx` and the user opens another editor for `xxx["aaa"]`,
        then to refresh the data of the second editor, the nested function
        `datafun` first gets the refreshed data for `xxx` and then gets the
        item with key "aaa".

        Parameters
        ----------
        index : QModelIndex
            Index of item whose current value is to be returned by the
            function constructed here.

        Returns
        -------
        Optional[Callable[[], Any]]
            Function which returns the current value of the data, or None if
            such a function cannot be constructed.
        """
        if self.data_function is None:
            return None
        key = index.model().keys[index.row()]

        def datafun():
            data = self.data_function()
            if isinstance(data, (tuple, list, dict, set)):
                return data[key]

            try:
                return getattr(data, key)
            except (NotImplementedError, AttributeError,
                    TypeError, ValueError):
                return None

        return datafun

    def show_warning(self, index):
        """
        Decide if showing a warning when the user is trying to view
        a big variable associated to a Tablemodel index.

        This avoids getting the variables' value to know its
        size and type, using instead those already computed by
        the TableModel.

        The problem is when a variable is too big, it can take a
        lot of time just to get its value.
        """
        val_type = index.sibling(index.row(), 1).data()
        val_size = index.sibling(index.row(), 2).data()

        if val_type in ['list', 'set', 'tuple', 'dict']:
            if int(val_size) > LARGE_COLLECTION:
                return True
        elif (val_type in ['DataFrame', 'Series'] or 'Array' in val_type or
                'Index' in val_type):
            # Avoid errors for user declared types that contain words like
            # the ones we're looking for above
            try:
                # From https://blender.stackexchange.com/a/131849
                shape = [int(s) for s in val_size.strip("()").split(",") if s]
                size = functools.reduce(operator.mul, shape)
                if size > LARGE_ARRAY:
                    return True
            except Exception:
                pass

        return False

    def createEditor(self, parent, option, index, object_explorer=False):
        """Overriding method createEditor"""
        self.sig_editor_creation_started.emit()
        if index.column() < 3:
            return None
        if self.show_warning(index):
            answer = QMessageBox.warning(
                self.parent(), _("Warning"),
                _("Opening this variable can be slow\n\n"
                  "Do you want to continue anyway?"),
                QMessageBox.Yes | QMessageBox.No)
            if answer == QMessageBox.No:
                self.sig_editor_shown.emit()
                return None
        try:
            value = self.get_value(index)
            if value is None:
                return None
        except Exception as exception:
            self.sig_editor_shown.emit()
            msg = _(
                "Spyder was unable to retrieve the value of this variable "
                "from the console."
            )
            self.show_error(exception, msg)
            return

        key = index.model().get_key(index)
        readonly = (isinstance(value, (tuple, set)) or self.parent().readonly
                    or not is_known_type(value))

        # We can't edit Numpy void objects because they could be anything, so
        # this might cause a crash.
        # Fixes spyder-ide/spyder#10603
        if isinstance(value, np.void):
            self.sig_editor_shown.emit()
            return None
        # CollectionsEditor for a list, tuple, dict, etc.
        elif (
            isinstance(value, (list, set, frozenset, tuple, dict))
            and not object_explorer
        ):
            from spyder.widgets.collectionseditor import CollectionsEditor
            editor = CollectionsEditor(
                parent=parent,
                namespacebrowser=self.namespacebrowser,
                data_function=self.make_data_function(index)
            )
            editor.setup(value, key, icon=self.parent().windowIcon(),
                         readonly=readonly)
            self.create_dialog(editor, dict(model=index.model(), editor=editor,
                                            key=key, readonly=readonly))
            return None
        # ArrayEditor for a Numpy array
        elif (isinstance(value, (np.ndarray, np.ma.MaskedArray)) and
                np.ndarray is not FakeObject and not object_explorer):
            # We need to leave this import here for tests to pass.
            from .arrayeditor import ArrayEditor
            editor = ArrayEditor(
                parent=parent,
                data_function=self.make_data_function(index)
            )
            if not editor.setup_and_check(value, title=key, readonly=readonly):
                self.sig_editor_shown.emit()
                return
            self.create_dialog(editor, dict(model=index.model(), editor=editor,
                                            key=key, readonly=readonly))
            return None
        # ArrayEditor for an images
        elif (isinstance(value, PIL.Image.Image) and
                np.ndarray is not FakeObject and
                PIL.Image is not FakeObject and
                not object_explorer):
            # Sometimes the ArrayEditor import above is not seen (don't know
            # why), so we need to reimport it here.
            # Fixes spyder-ide/spyder#16731
            from .arrayeditor import ArrayEditor
            arr = np.array(value)
            editor = ArrayEditor(parent=parent)
            if not editor.setup_and_check(arr, title=key, readonly=readonly):
                self.sig_editor_shown.emit()
                return
            conv_func = lambda arr: PIL.Image.fromarray(arr, mode=value.mode)
            self.create_dialog(editor, dict(model=index.model(), editor=editor,
                                            key=key, readonly=readonly,
                                            conv=conv_func))
            return None
        # DataFrameEditor for a pandas dataframe, series or index
        elif (isinstance(value, (pd.DataFrame, pd.Index, pd.Series))
                and pd.DataFrame is not FakeObject and not object_explorer):
            try:
                source_index = index.model().mapToSource(index)
            except AttributeError:
                # index.model() is not a QSortFilterProxyModel
                source_index = index

            type_of_value = source_index.model().row_type(source_index.row())
            if type_of_value == 'Polars DataFrame':
                readonly = True

            # We need to leave this import here for tests to pass.
            from .dataframeeditor import DataFrameEditor
            editor = DataFrameEditor(
                parent=parent,
                namespacebrowser=self.namespacebrowser,
                data_function=self.make_data_function(index),
                readonly=readonly
            )
            if not editor.setup_and_check(value, title=key):
                self.sig_editor_shown.emit()
                return
            self.create_dialog(editor, dict(model=index.model(), editor=editor,
                                            key=key, readonly=readonly))
            return None
        # QDateEdit and QDateTimeEdit for a dates or datetime respectively
        elif isinstance(value, datetime.date) and not object_explorer:
            if readonly:
                self.sig_editor_shown.emit()
                return None
            else:
                if isinstance(value, datetime.datetime):
                    editor = QDateTimeEdit(value, parent=parent)
                    # Needed to handle NaT values
                    # See spyder-ide/spyder#8329
                    try:
                        value.time()
                    except ValueError:
                        self.sig_editor_shown.emit()
                        return None
                else:
                    editor = QDateEdit(value, parent=parent)
                editor.setCalendarPopup(True)
                editor.setFont(
                    self.get_font(SpyderFontType.MonospaceInterface)
                )
                self.sig_editor_shown.emit()
                return editor
        # TextEditor for a long string
        elif (
            isinstance(value, (str, bytes))
            and len(value) > 40
            and not object_explorer
        ):
            te = TextEditor(None, parent=parent)
            if te.setup_and_check(value):
                editor = TextEditor(
                    value,
                    key,
                    readonly=True if isinstance(value, bytes) else readonly,
                    parent=parent,
                )
                self.create_dialog(
                    editor,
                    dict(
                        model=index.model(),
                        editor=editor,
                        key=key,
                        readonly=(
                            True if isinstance(value, bytes) else readonly
                        ),
                    ),
                )
            else:
                self.sig_editor_shown.emit()

            return None
        # QLineEdit for an individual value (int, float, short string, etc)
        elif is_editable_type(value) and not object_explorer:
            if readonly:
                self.sig_editor_shown.emit()
                return None
            elif (
                isinstance(value, int)
                and value.bit_length() >= MAX_INT_BIT_LENGTH_FOR_EDITING
            ):
                self.sig_editor_shown.emit()
                msg_box = QMessageBox(self.parent())
                msg_box.critical(
                    self.parent(),
                    _("Error"),
                    _("The value of this variable is too large to be edited.")
                )
                return None
            else:
                editor = QLineEdit(parent=parent)
                editor.setFont(
                    self.get_font(SpyderFontType.MonospaceInterface)
                )
                editor.setAlignment(Qt.AlignLeft)
                # This is making Spyder crash because the QLineEdit that it's
                # been modified is removed and a new one is created after
                # evaluation. So the object on which this method is trying to
                # act doesn't exist anymore.
                # editor.returnPressed.connect(self.commitAndCloseEditor)
                self.sig_editor_shown.emit()
                return editor
        # ObjectExplorer for an arbitrary Python object
        else:
            # Don't show the object explorer for short bytes because it's not
            # really necessary
            if isinstance(value, bytes):
                self.sig_editor_shown.emit()
                return None

            from spyder.plugins.variableexplorer.widgets.objectexplorer import (
                ObjectExplorer,
            )

            editor = ObjectExplorer(
                value,
                name=key,
                parent=parent,
                namespacebrowser=self.namespacebrowser,
                data_function=self.make_data_function(index),
                readonly=readonly)
            self.create_dialog(editor, dict(model=index.model(),
                                            editor=editor,
                                            key=key, readonly=readonly))
            return None

    def show_error(self, exception: Exception, msg: str) -> None:
        """
        Show error dialog box.

        This function is called when an error occurs while getting or setting
        a variable in the console. The dialog box has a "Show details" button,
        which shows the exception traceback. If an exception raised in the
        kernel is the cause, then also show that.

        Parameters
        ----------
        exception : Exception
            The error that occurred.
        msg : str
            Message informing the user of what Spyder was doing when the
            error occurred.
        """
        the_problem_is = _('The problem is:')
        contents = f'{msg}<br><br>{the_problem_is}<br>{exception}'
        details = ''.join(traceback.format_exception(exception))

        while exception:
            if (
                exception.args
                and isinstance(exception.args[0], CommsErrorWrapper)
            ):
                wrapper = exception.args[0]
                details += '\n'
                details += _('The following kernel error is associated:')
                details += '\n'
                details += ''.join(traceback.format_list(wrapper.tb))
                details += ''.join(
                    traceback.format_exception_only(wrapper.error)
                )
            exception = exception.__context__

        msg_box = QMessageBox(self.parent())
        msg_box.setTextFormat(Qt.RichText)  # Needed to enable links
        msg_box.setText(contents)
        msg_box.setDetailedText(details)
        msg_box.setWindowTitle(_('Error'))
        msg_box.setIcon(QMessageBox.Critical)
        msg_box.exec_()

    def create_dialog(self, editor, data):
        self._editors[id(editor)] = data
        editor.accepted.connect(
                     lambda eid=id(editor): self.editor_accepted(eid))
        editor.rejected.connect(
                     lambda eid=id(editor): self.editor_rejected(eid))
        self.sig_editor_shown.emit()
        editor.show()

    def editor_accepted(self, editor_id):
        data = self._editors[editor_id]
        if not data['readonly']:
            index = data['model'].get_index_from_key(data['key'])
            value = data['editor'].get_value()
            conv_func = data.get('conv', lambda v: v)

            try:
                self.set_value(index, conv_func(value))
            except Exception as exception:
                msg = _(
                    "Spyder was unable to set this variable in the console "
                    "to the new value."
                )
                self.show_error(exception, msg)

        # This is needed to avoid the problem reported on
        # spyder-ide/spyder#8557.
        try:
            self._editors.pop(editor_id)
        except KeyError:
            pass
        self.free_memory()

    def editor_rejected(self, editor_id):
        # This is needed to avoid the problem reported on
        # spyder-ide/spyder#8557.
        try:
            self._editors.pop(editor_id)
        except KeyError:
            pass
        self.free_memory()

    def free_memory(self):
        """Free memory after closing an editor."""
        try:
            self.sig_free_memory_requested.emit()
        except RuntimeError:
            pass

    def commitAndCloseEditor(self):
        """Overriding method commitAndCloseEditor"""
        editor = self.sender()
        # Avoid a segfault with PyQt5. Variable value won't be changed
        # but at least Spyder won't crash. It seems generated by a bug in sip.
        try:
            self.commitData.emit(editor)
        except AttributeError:
            pass
        self.closeEditor.emit(editor, QAbstractItemDelegate.NoHint)

    def setEditorData(self, editor, index):
        """
        Overriding method setEditorData
        Model --> Editor
        """
        value = self.get_value(index)
        if isinstance(editor, QLineEdit):
            if not isinstance(value, str):
                value = repr(value)
            editor.setText(value)
        elif isinstance(editor, QDateEdit):
            editor.setDate(value)
        elif isinstance(editor, QDateTimeEdit):
            editor.setDateTime(QDateTime(value.date(), value.time()))

    def setModelData(self, editor, model, index):
        """
        Overriding method setModelData
        Editor --> Model
        """
        if ((hasattr(model, "sourceModel")
                and not hasattr(model.sourceModel(), "set_value"))
                or not hasattr(model, "set_value")):
            # Read-only mode
            return

        if isinstance(editor, QLineEdit):
            value = editor.text()
            try:
                value = display_to_value(to_qvariant(value),
                                         self.get_value(index),
                                         ignore_errors=False)
            except Exception as msg:
                QMessageBox.critical(editor, _("Edit item"),
                                     _("<b>Unable to assign data to item.</b>"
                                       "<br><br>Error message:<br>%s"
                                       ) % str(msg))
                return
        elif isinstance(editor, QDateEdit):
            qdate = editor.date()
            value = datetime.date(qdate.year(), qdate.month(), qdate.day())
        elif isinstance(editor, QDateTimeEdit):
            qdatetime = editor.dateTime()
            qdate = qdatetime.date()
            qtime = qdatetime.time()
            # datetime uses microseconds, QDateTime returns milliseconds
            value = datetime.datetime(qdate.year(), qdate.month(), qdate.day(),
                                      qtime.hour(), qtime.minute(),
                                      qtime.second(), qtime.msec()*1000)
        else:
            # Should not happen...
            raise RuntimeError("Unsupported editor widget")
        self.set_value(index, value)

    def updateEditorGeometry(self, editor, option, index):
        """
        Overriding method updateEditorGeometry.

        This is necessary to set the correct position of the QLineEdit
        editor since option.rect doesn't have values -> QRect() and
        makes the editor to be invisible (i.e. it has 0 as x, y, width
        and height) when doing double click over a cell.
        See spyder-ide/spyder#9945
        """
        table_view = editor.parent().parent()
        if isinstance(table_view, QTableView):
            row = index.row()
            column = index.column()
            y0 = table_view.rowViewportPosition(row)
            x0 = table_view.columnViewportPosition(column)
            width = table_view.columnWidth(column)
            height = table_view.rowHeight(row)
            editor.setGeometry(x0, y0, width, height)
        else:
            super().updateEditorGeometry(editor, option, index)

    def paint(self, painter, option, index):
        """Actions to take when painting a cell."""
        if (
            # Do this only for the last column
            index.column() == 3
            # Do this when the row is hovered or if it's selected
            and (
                index.row() == self.parent().hovered_row
                or index.row() in self.parent().selected_rows()
            )
        ):
            # Paint regular contents
            super().paint(painter, option, index)

            # Paint an extra button to select the entire row. This is necessary
            # because in Spyder 6 is not intuitive how to do that since we use
            # a single click to open the editor associated to the cell.
            # Fixes spyder-ide/spyder#22524
            # Solution adapted from https://stackoverflow.com/a/11778012/438386

            # Getting the cell's rectangle
            rect = option.rect

            # Button left/top coordinates
            x = rect.left() + rect.width() - SELECT_ROW_BUTTON_SIZE
            y = rect.top() + rect.height() // 2 - SELECT_ROW_BUTTON_SIZE // 2

            # Create and paint button
            button = QStyleOptionButton()
            button.rect = QRect(
                x, y, SELECT_ROW_BUTTON_SIZE, SELECT_ROW_BUTTON_SIZE
            )
            button.text = ""
            button.icon = (
                ima.icon("select_row")
                if index.row() not in self.parent().selected_rows()
                else ima.icon("deselect_row")
            )
            button.iconSize = QSize(20, 20)
            button.state = QStyle.State_Enabled
            QApplication.style().drawControl(
                QStyle.CE_PushButtonLabel, button, painter
            )
        else:
            super().paint(painter, option, index)

    def editorEvent(self, event, model, option, index):
        """Actions to take when interacting with a cell."""
        if event.type() == QEvent.MouseButtonRelease and index.column() == 3:
            # Getting the position of the mouse click
            click_x = event.x()
            click_y = event.y()

            # Getting the cell's rectangle
            rect = option.rect

            # Region for the select row button
            x = rect.left() + rect.width() - SELECT_ROW_BUTTON_SIZE
            y = rect.top()

            # Select/deselect row when clicking on the button
            if click_x > x and (y < click_y < (y + SELECT_ROW_BUTTON_SIZE)):
                row = index.row()
                if row in self.parent().selected_rows():
                    # Deselect row if selected
                    index_left = index.sibling(row, 0)
                    index_right = index.sibling(row, 3)
                    selection = QItemSelection(index_left, index_right)
                    self.parent().selectionModel().select(
                        selection, QItemSelectionModel.Deselect
                    )
                else:
                    self.parent().selectRow(row)
            else:
                super().editorEvent(event, model, option, index)
        else:
            super().editorEvent(event, model, option, index)

        return False


class ToggleColumnDelegate(CollectionsDelegate):
    """ToggleColumn Item Delegate"""

    def __init__(self, parent=None, namespacebrowser=None,
                 data_function: Optional[Callable[[], Any]] = None):
        CollectionsDelegate.__init__(
            self, parent, namespacebrowser, data_function
        )
        self.current_index = None
        self.old_obj = None

    def restore_object(self):
        """Discart changes made to the current object in edition."""
        if self.current_index and self.old_obj is not None:
            index = self.current_index
            index.model().treeItem(index).obj = self.old_obj

    def get_value(self, index):
        """Get object value in index."""
        if index.isValid():
            value = index.model().treeItem(index).obj
            return value

    def set_value(self, index, value):
        if index.isValid():
            index.model().set_value(index, value)

    def make_data_function(
        self,
        index: QModelIndex
    ) -> Optional[Callable[[], Any]]:
        """
        Construct function which returns current value of data.

        This is used to refresh editors created from this piece of data.
        For instance, if `self` is the delegate for an editor displays the
        object `obj` and the user opens another editor for `obj.xxx.yyy`,
        then to refresh the data of the second editor, the nested function
        `datafun` first gets the refreshed data for `obj` and then gets the
        `xxx` attribute and then the `yyy` attribute.

        Parameters
        ----------
        index : QModelIndex
            Index of item whose current value is to be returned by the
            function constructed here.

        Returns
        -------
        Optional[Callable[[], Any]]
            Function which returns the current value of the data, or None if
            such a function cannot be constructed.
        """
        if self.data_function is None:
            return None

        obj_path = index.model().get_key(index).obj_path
        path_elements = obj_path.split('.')
        del path_elements[0]  # first entry is variable name

        def datafun():
            data = self.data_function()
            try:
                for attribute_name in path_elements:
                    data = getattr(data, attribute_name)
                return data
            except (NotImplementedError, AttributeError,
                    TypeError, ValueError):
                return None

        return datafun

    def createEditor(self, parent, option, index):
        """Overriding method createEditor"""
        if self.show_warning(index):
            answer = QMessageBox.warning(
                self.parent(), _("Warning"),
                _("Opening this variable can be slow\n\n"
                  "Do you want to continue anyway?"),
                QMessageBox.Yes | QMessageBox.No)
            if answer == QMessageBox.No:
                return None
        try:
            value = self.get_value(index)
            try:
                self.old_obj = value.copy()
            except AttributeError:
                self.old_obj = copy.deepcopy(value)
            if value is None:
                return None
        except Exception as msg:
            msg_box = QMessageBox(self.parent())
            msg_box.setTextFormat(Qt.RichText)  # Needed to enable links
            msg_box.critical(
                self.parent(),
                _("Error"),
                _(
                    "Spyder was unable to retrieve the value of this variable "
                    "from the console.<br><br>"
                    "The problem is:<br>"
                    "<i>%s</i>"
                ) % str(msg)
            )
            return
        self.current_index = index

        key = index.model().get_key(index).obj_name
        readonly = (isinstance(value, (tuple, set)) or self.parent().readonly
                    or not is_known_type(value))

        # CollectionsEditor for a list, tuple, dict, etc.
        if isinstance(value, (list, set, frozenset, tuple, dict)):
            from spyder.widgets.collectionseditor import CollectionsEditor
            editor = CollectionsEditor(
                parent=parent,
                namespacebrowser=self.namespacebrowser,
                data_function=self.make_data_function(index)
            )
            editor.setup(value, key, icon=self.parent().windowIcon(),
                         readonly=readonly)
            self.create_dialog(editor, dict(model=index.model(), editor=editor,
                                            key=key, readonly=readonly))
            return None
        # ArrayEditor for a Numpy array
        elif (isinstance(value, (np.ndarray, np.ma.MaskedArray)) and
                np.ndarray is not FakeObject):
            editor = ArrayEditor(
                parent=parent,
                data_function=self.make_data_function(index)
            )
            if not editor.setup_and_check(value, title=key, readonly=readonly):
                return
            self.create_dialog(editor, dict(model=index.model(), editor=editor,
                                            key=key, readonly=readonly))
            return None
        # ArrayEditor for an images
        elif (isinstance(value, PIL.Image.Image) and
                np.ndarray is not FakeObject and PIL.Image is not FakeObject):
            arr = np.array(value)
            editor = ArrayEditor(parent=parent)
            if not editor.setup_and_check(arr, title=key, readonly=readonly):
                return
            conv_func = lambda arr: PIL.Image.fromarray(arr, mode=value.mode)
            self.create_dialog(editor, dict(model=index.model(), editor=editor,
                                            key=key, readonly=readonly,
                                            conv=conv_func))
            return None
        # DataFrameEditor for a pandas dataframe, series or index
        elif (isinstance(value, (pd.DataFrame, pd.Index, pd.Series))
                and pd.DataFrame is not FakeObject):
            editor = DataFrameEditor(
                parent=parent,
                namespacebrowser=self.namespacebrowser,
                data_function=self.make_data_function(index)
            )
            if not editor.setup_and_check(value, title=key):
                return
            self.create_dialog(editor, dict(model=index.model(), editor=editor,
                                            key=key, readonly=readonly))
            return None
        # QDateEdit and QDateTimeEdit for a dates or datetime respectively
        elif isinstance(value, datetime.date):
            if readonly:
                return None
            else:
                if isinstance(value, datetime.datetime):
                    editor = QDateTimeEdit(value, parent=parent)
                else:
                    editor = QDateEdit(value, parent=parent)
                editor.setCalendarPopup(True)
                editor.setFont(
                    self.get_font(SpyderFontType.MonospaceInterface)
                )
                return editor
        # TextEditor for a long string
        elif isinstance(value, (str, bytes)) and len(value) > 40:
            te = TextEditor(None, parent=parent)
            if te.setup_and_check(value):
                editor = TextEditor(
                    value,
                    key,
                    readonly=True if isinstance(value, bytes) else readonly,
                    parent=parent,
                )
                self.create_dialog(
                    editor,
                    dict(
                        model=index.model(),
                        editor=editor,
                        key=key,
                        readonly=(
                            True if isinstance(value, bytes) else readonly
                        ),
                    ),
                )
            return None
        # QLineEdit for an individual value (int, float, short string, etc)
        elif is_editable_type(value):
            if readonly:
                return None
            else:
                editor = QLineEdit(parent=parent)
                editor.setFont(
                    self.get_font(SpyderFontType.MonospaceInterface)
                )
                editor.setAlignment(Qt.AlignLeft)
                # This is making Spyder crash because the QLineEdit that it's
                # been modified is removed and a new one is created after
                # evaluation. So the object on which this method is trying to
                # act doesn't exist anymore.
                # editor.returnPressed.connect(self.commitAndCloseEditor)
                return editor
        # An arbitrary Python object.
        # Since we are already in the Object Explorer no editor is needed
        else:
            return None

    def editor_accepted(self, editor_id):
        """Actions to execute when the editor has been closed."""
        data = self._editors[editor_id]
        if not data['readonly'] and self.current_index:
            index = self.current_index
            value = data['editor'].get_value()
            conv_func = data.get('conv', lambda v: v)
            self.set_value(index, conv_func(value))
        # This is needed to avoid the problem reported on
        # spyder-ide/spyder#8557.
        try:
            self._editors.pop(editor_id)
        except KeyError:
            pass
        self.free_memory()

    def editor_rejected(self, editor_id):
        """Actions to do when the editor was rejected."""
        self.restore_object()
        super().editor_rejected(editor_id)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

from __future__ import annotations
import asyncio
from enum import Enum
import fnmatch
import functools
import io
import logging
import os
import posixpath
from datetime import datetime

from aiohttp.client_exceptions import ClientResponseError
from qtpy.compat import getexistingdirectory, getopenfilenames
from qtpy.QtCore import QSortFilterProxyModel, Qt, Signal
from qtpy.QtGui import (
    QClipboard,
    QKeySequence,
    QStandardItem,
    QStandardItemModel,
)
from qtpy.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QInputDialog,
    QMessageBox,
    QLineEdit,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

from spyder.api.asyncdispatcher import AsyncDispatcher
from spyder.api.config.decorators import on_conf_change
from spyder.api.translations import _
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.config.base import get_conf_path
from spyder.config.utils import EDIT_EXTENSIONS
from spyder.plugins.editor.utils.editor import get_default_file_content
from spyder.plugins.remoteclient.api.modules.base import (
    SpyderRemoteSessionClosed,
)
from spyder.plugins.remoteclient.api.protocol import ClientType
from spyder.plugins.remoteclient.api.modules.file_services import (
    RemoteFileServicesError,
    RemoteOSError,
    SpyderRemoteFileServicesAPI,
)
from spyder.plugins.shortcuts.utils import (
    ShortcutData,
    SHORTCUTS_FOR_WIDGETS_DATA,
)
from spyder.utils.icon_manager import ima
from spyder.utils.misc import getcwd_or_home
from spyder.utils.qthelpers import keyevent_to_keysequence_str


logger = logging.getLogger(__name__)


class RemoteViewMenus:
    Context = "remote_context_menu"
    New = "remote_new_menu"


class RemoteViewNewSubMenuSections:
    General = 'remote_general_section'
    Language = 'remote_language_section'


class RemoteExplorerActions:
    Copy = "remote_copy_action"
    Paste = "remote_paste_action"
    CopyPath = "remote_copy_path_action"
    Delete = "remote_delete_action"
    Download = "remote_download_action"
    NewDirectory = "remote_new_directory_action"
    NewFile = "remote_new_file_action"
    NewModule = "remote_new_module_action"
    NewPackage = "remote_new_package_action"
    Rename = "remote_rename_action"
    Upload = "remote_upload_file_action"


class RemoteExplorerContextMenuSections:
    CopyPaste = "remote_copy_paste_section"
    Extras = "remote_extras_section"
    New = "remote_new_section"


class RemoteExistenceOperations(Enum):
    Upload = "upload"
    Paste = "paste"
    NewFile = "new_file"
    NewFileWithContent = "new_file_with_content"
    NewDirectoryWithContent = "new_directory_with_content"


class RemoteQSortFilterProxyModel(QSortFilterProxyModel):

    def lessThan(self, left, right):
        right_data = self.sourceModel().data(
            self.sourceModel().index(right.row(), 0), Qt.UserRole + 1
        )
        if right_data["type"] == "ACTION":
            return self.sortOrder() == Qt.AscendingOrder

        left_data = self.sourceModel().data(
            self.sourceModel().index(left.row(), 0), Qt.UserRole + 1
        )
        if left_data["type"] == "ACTION":
            return self.sortOrder() == Qt.DescendingOrder

        if left_data["type"] == "directory" and right_data["type"] == "file":
            return True

        if left_data["type"] == "file" and right_data["type"] == "directory":
            return False

        return super().lessThan(left, right)


class RemoteExplorer(QWidget, SpyderWidgetMixin):

    sig_dir_opened = Signal(str, str)
    sig_start_spinner_requested = Signal()
    sig_stop_spinner_requested = Signal()

    def __init__(self, parent=None, class_parent=None, files=None):
        super().__init__(parent=parent, class_parent=parent)

        # General attributes
        self.remote_files_manager: SpyderRemoteFileServicesAPI | None = None
        self.server_id: str | None = None
        self.root_prefix: dict[str, str] = {}

        self.background_files_load = set()
        self.extra_files = []
        self.more_files_available = False

        self.filter_on = False
        self.name_filters = []

        self.history = []
        self.histindex = None

        self._files_to_copy: dict[str, list[str]] = {}
        self._files_to_delete: dict[str, int] = {}
        self._files_to_download: dict[str, int] = {}
        self._files_to_paste: dict[str, int] = {}
        self._files_to_rename: dict[str, int] = {}
        self._files_to_upload: dict[str, int] = {}

        # Model, actions and widget setup
        self.context_menu = self.create_menu(RemoteViewMenus.Context)
        new_submenu = self.create_menu(
            RemoteViewMenus.New,
            _('New'),
        )

        self.new_package_action = self.create_action(
            RemoteExplorerActions.NewPackage,
            text=_("Python package..."),
            icon=self.create_icon('package_new'),
            triggered=self.new_package,
        )
        self.new_module_action = self.create_action(
            RemoteExplorerActions.NewModule,
            text=_("Python file..."),
            icon=self.create_icon('python'),
            triggered=self.new_module,
        )
        self.new_directory_action = self.create_action(
            RemoteExplorerActions.NewDirectory,
            text=_("Folder..."),
            icon=self.create_icon('folder_new'),
            triggered=self.new_directory,
        )
        self.new_file_action = self.create_action(
            RemoteExplorerActions.NewFile,
            text=_("File..."),
            icon=self.create_icon('TextFileIcon'),
            triggered=self.new_file,
        )
        self.copy_action = self.create_action(
            RemoteExplorerActions.Copy,
            _("Copy"),
            icon=self.create_icon("editcopy"),
            triggered=self.copy_items,
        )
        self.paste_action = self.create_action(
            RemoteExplorerActions.Paste,
            _("Paste"),
            icon=self.create_icon("editpaste"),
            triggered=self.paste_items,
        )
        self.rename_action = self.create_action(
            RemoteExplorerActions.Rename,
            _("Rename..."),
            icon=self.create_icon("rename"),
            triggered=self.rename_items,
        )
        self.copy_path_action = self.create_action(
            RemoteExplorerActions.CopyPath,
            _("Copy absolute path"),
            triggered=self.copy_paths,
        )
        self.delete_action = self.create_action(
            RemoteExplorerActions.Delete,
            _("Delete..."),
            icon=self.create_icon("editclear"),
            triggered=self.delete_items,
        )
        self.download_action = self.create_action(
            RemoteExplorerActions.Download,
            _("Download..."),
            icon=self.create_icon("fileimport"),
            triggered=self.download_items,
        )
        self.upload_file_action = self.create_action(
            RemoteExplorerActions.Upload,
            _("Upload files"),
            icon=self.create_icon("fileexport"),
            triggered=self.upload_files,
        )

        for item in [self.new_file_action, self.new_directory_action]:
            self.add_item_to_menu(
                item,
                new_submenu,
                section=RemoteViewNewSubMenuSections.General,
            )

        for item in [self.new_module_action, self.new_package_action]:
            self.add_item_to_menu(
                item,
                new_submenu,
                section=RemoteViewNewSubMenuSections.Language,
            )

        for item in [
            new_submenu,
            self.rename_action,
            self.delete_action,
        ]:
            self.add_item_to_menu(
                item,
                self.context_menu,
                section=RemoteExplorerContextMenuSections.New,
            )

        for item in [
            self.copy_action,
            self.paste_action,
            self.copy_path_action,
        ]:
            self.add_item_to_menu(
                item,
                self.context_menu,
                section=RemoteExplorerContextMenuSections.CopyPaste,
            )

        for item in [self.upload_file_action, self.download_action]:
            self.add_item_to_menu(
                item,
                self.context_menu,
                section=RemoteExplorerContextMenuSections.Extras,
            )

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_show_context_menu)

        self._register_shortcuts()

        self.model = QStandardItemModel(self)
        self.model.setHorizontalHeaderLabels(
            [
                "Name",
                "Size",
                "Type",
                "Date Modified",
            ]
        )
        self.proxy_model = RemoteQSortFilterProxyModel(self)
        self.proxy_model.setSourceModel(self.model)

        self.view = QTreeView(self)
        self.view.setModel(self.proxy_model)
        self.view.setSortingEnabled(True)
        self.view.setSelectionMode(QTreeView.ExtendedSelection)
        self.view.setSelectionBehavior(
            QAbstractItemView.SelectionBehavior.SelectRows
        )
        self.view.setContextMenuPolicy(Qt.CustomContextMenu)
        self.view.customContextMenuRequested.connect(
            self._on_show_context_menu
        )

        if files:
            self.set_files(files)

        self.view.sortByColumn(0, Qt.AscendingOrder)
        self.view.entered.connect(self._on_entered_item)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.view)

    @on_conf_change(
        option=[
            "size_column",
            "type_column",
            "date_column",
            "name_filters",
            "show_hidden",
            "single_click_to_open",
        ]
    )
    def on_conf_update(self, option, value):
        if option == "size_column":
            self.view.setColumnHidden(1, not value)
        elif option == "type_column":
            self.view.setColumnHidden(2, not value)
        elif option == "date_column":
            self.view.setColumnHidden(3, not value)
        elif option == "name_filters":
            if self.filter_on:
                self.filter_files(value)
        elif option == "show_hidden":
            self.refresh(force_current=True)
        elif option == "single_click_to_open":
            self.set_single_click_to_open(value)

    # ---- Private API
    # -------------------------------------------------------------------------
    def _on_show_context_menu(self, position):
        index = self.view.indexAt(position)
        if not index.isValid():
            self.view.setCurrentIndex(index)
            self.view.clearSelection()
            data = {}
            data_available = False
            is_file = False
        else:
            source_index = self.proxy_model.mapToSource(
                self.view.currentIndex()
            )
            data_index = self.model.index(source_index.row(), 0)
            data = self.model.data(data_index, Qt.UserRole + 1)
            data_available = bool(data)
            is_file = (
                data_available and data.get("type", "directory") == "file"
            )

        # Disable actions that require a valid selection
        self.rename_action.setEnabled(data_available)
        self.delete_action.setEnabled(data_available)

        # Disable actions not suitable for directories
        self.copy_action.setEnabled(is_file)
        self.paste_action.setEnabled(
            bool(self._files_to_copy.get(self.server_id, []))
        )

        global_position = self.mapToGlobal(position)
        self.context_menu.popup(global_position)

    def _on_entered_item(self, index):
        if self.get_conf("single_click_to_open"):
            self.view.setCursor(Qt.PointingHandCursor)
            self.view.header().setCursor(Qt.ArrowCursor)
        else:
            self.view.setCursor(Qt.ArrowCursor)

    def _on_clicked_item(self, index):
        source_index = self.proxy_model.mapToSource(index)
        data_index = self.model.index(source_index.row(), 0)
        data = self.model.data(data_index, Qt.UserRole + 1)
        if data:
            data_name = data["name"]
            data_type = data["type"]
            if data_type == "directory":
                self.chdir(data_name, emit=True)
            elif (
                data_type == "file"
                and os.path.splitext(data_name)[1] in EDIT_EXTENSIONS
            ):
                QMessageBox.warning(
                    self,
                    _("Unsupported functionality"),
                    _(
                        "Opening remote files in the Editor is not yet "
                        "supported. This feature will be added in Spyder "
                        "6.2.0"
                    )
                )
            elif data_type == "ACTION" and data_name == "FETCH_MORE":
                self.fetch_more_files()

    @property
    def _operation_in_progress(self):
        return (
            self._files_to_delete.get(self.server_id, 0) > 0
            or self._files_to_download.get(self.server_id, 0) > 0
            or self._files_to_paste.get(self.server_id, 0) > 0
            or self._files_to_rename.get(self.server_id, 0) > 0
            or self._files_to_upload.get(self.server_id, 0) > 0
        )

    def _handle_future_response_error(
        self, response, error_title, error_message
    ):
        # Catch errors raised when running async methods.
        # Fixes spyder-ide/spyder#24974
        try:
            response.result()
            return False
        except (RemoteOSError, OSError) as error:
            logger.debug(error)

            if hasattr(error, "message"):
                str_error = error.message
            else:
                str_error = str(error)

            error_message = f"{error_message}<br><br>{str_error}"
            QMessageBox.critical(self, error_title, error_message)
            return True

    @AsyncDispatcher.QtSlot
    def _on_remote_new_package(self, future, package_name):
        error = self._handle_future_response_error(
            future,
            _("New Python Package error"),
            _("An error occured while trying to create a new Python package"),
        )

        if not error:
            # Only create init file if there weren't errors creating its
            # parent.
            new_name = posixpath.join(package_name, "__init__.py")
            self._new_item(new_name, for_file=True, with_content=True)
        else:
            if not self._operation_in_progress:
                self.sig_stop_spinner_requested.emit()

    @AsyncDispatcher.QtSlot
    def _on_remote_new_module(self, future):
        self._handle_future_response_error(
            future,
            _("New Python File error"),
            _("An error occured while trying to create a new Python file"),
        )
        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_new_module(self, new_path, file_content):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        async with await self.remote_files_manager.open(
            new_path, mode="w"
        ) as file_manager:
            return await file_manager.write(file_content)

    @AsyncDispatcher.QtSlot
    def _on_remote_new(self, future):
        self._handle_future_response_error(
            future,
            _("New error"),
            _("An error occured while trying to create a file/directory"),
        )
        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_new(
        self, new_path, for_file=False
    ):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        if for_file:
            response = await self.remote_files_manager.touch(new_path)
        else:
            response = await self.remote_files_manager.mkdir(new_path)

        return response

    @AsyncDispatcher.QtSlot
    def _on_remote_paste(self, future):
        self._handle_future_response_error(
            future,
            _("Paste error"),
            _(
                "An error occured while trying to paste a file or directory"
            ),
        )

        if self._files_to_paste[self.server_id] > 0:
            self._files_to_paste[self.server_id] -= 1

        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_paste(self, old_path, new_path):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        return await self.remote_files_manager.copy(old_path, new_path)

    @AsyncDispatcher.QtSlot
    def _on_remote_rename(self, future):
        self._handle_future_response_error(
            future,
            _("Rename error"),
            _("An error occured while trying to rename a file"),
        )

        if self._files_to_rename[self.server_id] > 0:
            self._files_to_rename[self.server_id] -= 1

        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_rename(self, old_path, new_path):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        return await self.remote_files_manager.replace(old_path, new_path)

    @AsyncDispatcher.QtSlot
    def _on_remote_delete(self, future):
        self._handle_future_response_error(
            future,
            _("Delete error"),
            _("An error occured while trying to delete a file/directory"),
        )

        if self._files_to_delete[self.server_id] > 0:
            self._files_to_delete[self.server_id] -= 1

        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_delete(self, path, is_file=False):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        if is_file:
            response = await self.remote_files_manager.unlink(path)
        else:
            response = await self.remote_files_manager.rmdir(
                path, non_empty=True
            )

        return response

    @AsyncDispatcher.QtSlot
    def _on_remote_download_file(
        self, future, remote_filename, local_directory, is_file
    ):
        download_error = False
        try:
            data = future.result()
        except (RemoteFileServicesError, ClientResponseError) as error:
            download_error = True
            logger.debug(error)

            if is_file:
                message = _(
                    "An error occurred while trying to download the file "
                    "<b>{}</b> from server <b>{}</b>:"
                    "<br><br>"
                    "{}"
                ).format(
                    remote_filename, self._get_server_name(), error.message
                )
            else:
                remote_dir = remote_filename.split(".")[0]

                if isinstance(error, ClientResponseError):
                    message = _(
                        "It seems you don't have permissions to read the "
                        "directory <b>{}</b> in server <b>{}</b>."
                    ).format(remote_dir, self._get_server_name())
                else:
                    message = _(
                        "An error occurred while trying to download the "
                        "directory <b>{}</b> as a zip file from server "
                        "<b>{}</b>:"
                        "<br><br>"
                        "{}"
                    ).format(
                        remote_dir, self._get_server_name(), error.message
                    )

        if download_error:
            QMessageBox.critical(self, _("Download error"), message)
        elif os.path.isdir(local_directory):
            # Local filename
            local_filename = os.path.join(local_directory, remote_filename)

            # Write remote contents to local file
            try:
                with open(local_filename, "w") as download_file:
                    download_file.write(data)
            except TypeError:
                with open(local_filename, "wb") as download_file:
                    download_file.write(data)

        if self._files_to_download[self.server_id] > 0:
            self._files_to_download[self.server_id] -= 1

        if not self._operation_in_progress:
            self.sig_stop_spinner_requested.emit()

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_download_directory(self, path):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        zip_generator = self.remote_files_manager.zip_directory(path)
        zip_data = io.BytesIO()
        async for data in zip_generator:
            zip_data.write(data)
        zip_data.seek(0)
        return zip_data.getbuffer()

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_download_file(self, path):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        try:
            async with await self.remote_files_manager.open(
                path, mode="r"
            ) as file_manager:
                return await file_manager.read()
        except RemoteFileServicesError:
            async with await self.remote_files_manager.open(
                path, mode="rb"
            ) as file_manager:
                return await file_manager.read()

    @AsyncDispatcher.QtSlot
    def _on_remote_upload_file(self, future):
        self._handle_future_response_error(
            future,
            _("Upload error"),
            _("An error occured while trying to upload a file"),
        )

        if self._files_to_upload[self.server_id] > 0:
            self._files_to_upload[self.server_id] -= 1

        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_upload_file(self, local_path):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        file_content = None
        remote_file = posixpath.join(
            self.root_prefix[self.server_id], os.path.basename(local_path)
        )

        try:
            with open(local_path, mode="r") as local_file:
                file_content = local_file.read()
            mode = "w"
        except UnicodeDecodeError:
            with open(local_path, mode="rb") as local_file:
                file_content = local_file.read()
            mode="wb"

        if file_content:
            async with await self.remote_files_manager.open(
                remote_file, mode=mode
            ) as file_manager:
                return await file_manager.write(file_content)

    @AsyncDispatcher.QtSlot
    def _on_remote_ls(self, future):
        error = self._handle_future_response_error(
            future,
            _("List contents error"),
            _("An error occured while trying to view a directory"),
        )

        if not error:
            data = future.result()
            self.set_files(data)
        else:
            # Set parent directory as root because the selected one can't be
            # accessed
            self.chdir(os.path.dirname(self.root_prefix[self.server_id]))

        if not self._operation_in_progress:
            self.sig_stop_spinner_requested.emit()

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_ls(self, path, server_id):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        for task in self.background_files_load:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

        self.extra_files = []
        self.more_files_available = False
        files = []
        try:
            init_files_display = self.get_conf("init_files_display")
            generator = self.remote_files_manager.ls(path)
            async for file in generator:
                file_name = os.path.relpath(
                    file["name"], self.root_prefix[self.server_id]
                )
                file_type = file["type"]
                if len(files) < init_files_display:
                    if not self.get_conf(
                        "show_hidden"
                    ) and file_name.startswith("."):
                        continue
                    if (
                        self.name_filters
                        and len(self.name_filters)
                        and file_type == "file"
                    ):
                        for name_filter in self.name_filters:
                            if file_type == "file" and fnmatch.fnmatch(
                                file_name, name_filter
                            ):
                                files.append(file)
                                break
                    else:
                        files.append(file)
                else:
                    break

            if len(files) == init_files_display:
                task = asyncio.create_task(
                    self._get_extra_files(generator, len(files))
                )
                self.background_files_load.add(task)
                task.add_done_callback(self.background_files_load.discard)
        except SpyderRemoteSessionClosed:
            self.remote_files_manager = None

        return files

    def _on_check_if_remote_files_exist(
        self, future, operation: RemoteExistenceOperations
    ):
        self._handle_future_response_error(
            future,
            _("Upload error"),
            _("An error occured while trying to upload files"),
        )

        yes_to_all = False
        paths_existence = future.result()

        if len(paths_existence) > 1:
            buttons = (
                QMessageBox.Yes
                | QMessageBox.YesToAll
                | QMessageBox.No
                | QMessageBox.Cancel
            )
        else:
            buttons = QMessageBox.Yes | QMessageBox.No

        for path, response in paths_existence:
            filename = os.path.basename(path)
            files_counter = None

            if not yes_to_all and response["exists"]:
                if operation == RemoteExistenceOperations.Paste:
                    opening_sentence = _(
                        "The file <b>{}</b> that you're trying to paste on "
                        "the server <b>{}</b> already exists in the current "
                        "location."
                    )

                    files_counter = self._files_to_paste
                elif operation == RemoteExistenceOperations.Upload:
                    opening_sentence = _(
                        "The file <b>{}</b> that you're trying to upload to "
                        "the server <b>{}</b> already exists in the current "
                        "location."
                    )

                    files_counter = self._files_to_upload
                elif operation in [
                    RemoteExistenceOperations.NewFile,
                    RemoteExistenceOperations.NewFileWithContent,
                ]:
                    opening_sentence = _(
                        "The file <b>{}</b> that you're trying to create on "
                        "the server <b>{}</b> already exists in the current "
                        "location."
                    )
                elif (
                    operation
                    == RemoteExistenceOperations.NewDirectoryWithContent
                ):
                    # Show the same message as the one for empty directories.
                    opening_sentence = _(
                        "An error occurred while trying to create a file/"
                        "directory"
                    )
                    msg = (
                        opening_sentence
                        + "<br><br>"
                        + "[Errno 17] File exists: '{}'".format(path)
                    )

                    QMessageBox.critical(self, _("Error"), msg)
                    self.sig_stop_spinner_requested.emit()
                    return

                msg = (
                    opening_sentence.format(filename, self._get_server_name())
                    + "<br><br>"
                    + _("Do you want to overwrite it?")
                )
                answer = QMessageBox.warning(
                    self,
                    _("Overwrite file"),
                    msg,
                    buttons,
                )

                if answer == QMessageBox.YesToAll:
                    yes_to_all = True
                elif answer == QMessageBox.No:
                    if (
                        files_counter is not None
                        and files_counter[self.server_id] > 0
                    ):
                        files_counter[self.server_id] -= 1

                    if not self._operation_in_progress:
                        self.sig_stop_spinner_requested.emit()

                    continue
                elif answer == QMessageBox.Cancel:
                    if files_counter is not None:
                        files_counter[self.server_id] = 0

                    if not self._operation_in_progress:
                        self.sig_stop_spinner_requested.emit()

                    return

            if operation == RemoteExistenceOperations.Paste:
                new_path = posixpath.join(
                    self.root_prefix[self.server_id], filename
                )
                self._do_remote_paste(path, new_path).connect(
                    self._on_remote_paste
                )
            elif operation == RemoteExistenceOperations.Upload:
                self._do_remote_upload_file(path).connect(
                    self._on_remote_upload_file
                )
            elif operation == RemoteExistenceOperations.NewFile:
                self._do_remote_new(path, for_file=True).connect(
                    self._on_remote_new
                )
            elif operation == RemoteExistenceOperations.NewFileWithContent:
                file_content, __, __ = get_default_file_content(
                    get_conf_path("template.py")
                )
                self._do_remote_new_module(path, file_content).connect(
                    self._on_remote_new_module
                )
            elif (
                operation == RemoteExistenceOperations.NewDirectoryWithContent
            ):
                @AsyncDispatcher.QtSlot
                def remote_package(future):
                    self._on_remote_new_package(future, filename)

                self._do_remote_new(path).connect(remote_package)

    @AsyncDispatcher(loop="explorer")
    async def _check_if_remote_files_exist(self, paths: list[str]):
        """Check if remote files exist in the remote cwd."""
        paths_existence = []

        for path in paths:
            remote_file = posixpath.join(
                self.root_prefix[self.server_id], os.path.basename(path)
            )

            paths_existence.append(
                (path, await self.remote_files_manager.exists(remote_file))
            )

        return paths_existence

    async def _get_extra_files(self, generator, already_added):
        self.extra_files = []
        self.more_files_available = False
        async for file in generator:
            if len(self.extra_files) + already_added == self.get_conf(
                "max_files_display"
            ):
                self.more_files_available = True
                break

            file_name = os.path.relpath(
                file["name"], self.root_prefix[self.server_id]
            )
            file_type = file["type"]
            if not self.get_conf("show_hidden") and file_name.startswith("."):
                continue

            if (
                self.name_filters
                and len(self.name_filters)
                and file_type == "file"
            ):
                for name_filter in self.name_filters:
                    if file_type == "file" and fnmatch.fnmatch(
                        file_name, name_filter
                    ):
                        self.extra_files.append(file)
                        break
            else:
                self.extra_files.append(file)

        logger.debug(
            f"{len(self.extra_files)} available extra files to be shown"
        )

    def _new_item(self, new_name, for_file=False, with_content=False):
        new_path = posixpath.join(self.root_prefix[self.server_id], new_name)

        if for_file:
            # Check if remote file with the same name exists before trying to
            # create it.
            self._check_if_remote_files_exist([new_path]).connect(
                AsyncDispatcher.QtSlot(
                    functools.partial(
                        self._on_check_if_remote_files_exist,
                        operation=(
                            RemoteExistenceOperations.NewFileWithContent
                            if with_content
                            else RemoteExistenceOperations.NewFile
                        ),
                    )
                )
            )
        else:
            if with_content:
                # This is necessary because although an existing directory is
                # not overwritten by the files API, an __init__.py file inside
                # it is.
                self._check_if_remote_files_exist([new_path]).connect(
                    AsyncDispatcher.QtSlot(
                        functools.partial(
                            self._on_check_if_remote_files_exist,
                            operation=(
                                RemoteExistenceOperations.NewDirectoryWithContent
                            ),
                        )
                    )
                )
            else:
                # No need to check for existence because the files API raises
                # an error.
                self._do_remote_new(new_path, for_file=False).connect(
                    self._on_remote_new
                )

        self.sig_start_spinner_requested.emit()

    def _get_selected_indexes(self):
        indexes = self.view.selectedIndexes()
        if not indexes:
            return []

        # Indexes corresponding to columns other than the first are not
        # necessary because the info we need for multi-file operations is
        # available in the first one.
        return [idx for idx in indexes if idx.column() == 0]

    def _get_server_name(self):
        if (
            self.get_conf(
                f"{self.server_id}/client_type", section="remoteclient"
            )
            == ClientType.SSH
        ):
            auth_method = self.get_conf(
                f"{self.server_id}/auth_method", section="remoteclient"
            )
            name = self.get_conf(
                f"{self.server_id}/{auth_method}/name", section="remoteclient"
            )
        else:
            name = self.get_conf(
                f"{self.server_id}/jupyterhub_login/name",
                section="remoteclient",
            )

        return name

    def _register_shortcut_for_action(self, shortcut, action):
        action.setShortcut(QKeySequence(shortcut))

    def _register_shortcuts(self):
        """
        Register shortcuts for several actions.

        We have to do this to use the same shortcuts as the ones set for the
        local explorer without adding to Preferences more entries with
        different ids for the same actions.
        """
        for name, action in [
            ("copy file", self.copy_action),
            ("paste file", self.paste_action),
            ("copy absolute path", self.copy_path_action),
        ]:
            # This is necessary so that the shortcut options emit a
            # notification when changed in Preferneces.
            data = ShortcutData(
                qobject=None,
                name=name,
                context=self.CONF_SECTION,
            )
            SHORTCUTS_FOR_WIDGETS_DATA.append(data)

            # Add observer for shortcut to update it in its corresponding
            # action
            self.add_configuration_observer(
                functools.partial(
                    self._register_shortcut_for_action,
                    action=action,
                ),
                option=f"{self.CONF_SECTION}/{name}",
                section="shortcuts",
            )

    def _format_file_size(self, size):
        """
        Format file size using the same format used by Qt for local files.

        Adapted from https://stackoverflow.com/a/58201995/438386
        """
        units = ['bytes', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', "EiB", "ZiB"]
        index = 0

        if size > 0:
            while size >= 1024:
                size /= 1024
                index += 1

            formatted_size = f"{size:.2f}"
        else:
            formatted_size = "0"

        return f'{formatted_size} {units[index]}'

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def keyPressEvent(self, event):
        """Handle keyboard shortcuts and special keys."""
        key_seq = keyevent_to_keysequence_str(event)

        if event.key() == Qt.Key_F2:
            self.rename_items()
        elif event.key() == Qt.Key_Delete:
            self.delete_items()
        elif event.key() == Qt.Key_Backspace:
            self.go_to_parent_directory()
        elif key_seq == self.copy_action.shortcut().toString():
            self.copy_items()
        elif key_seq == self.paste_action.shortcut().toString():
            self.paste_items()
        elif key_seq == self.copy_path_action.shortcut().toString():
            self.copy_paths()
        else:
            super().keyPressEvent(event)

    # ---- Public API
    # -------------------------------------------------------------------------
    @AsyncDispatcher.QtSlot
    def chdir(
        self,
        directory=None,
        server_id=None,
        emit=True,
        browsing_history=False,
        remote_files_manager=None,
    ):
        if browsing_history:
            directory = self.history[self.histindex]
        elif directory in self.history:
            self.histindex = self.history.index(directory)
        else:
            if self.histindex is None:
                self.history = []
            else:
                self.history = self.history[: self.histindex + 1]
            if len(self.history) == 0 or (
                self.history and self.history[-1] != directory
            ):
                self.history.append(directory)
            self.histindex = len(self.history) - 1

        if directory == self.root_prefix.get(server_id):
            return

        if server_id:
            self.server_id = server_id
        self.root_prefix[self.server_id] = directory
        if remote_files_manager:
            self.remote_files_manager = remote_files_manager
        self.refresh(force_current=True)
        if emit:
            self.sig_dir_opened.emit(directory, self.server_id)

    def set_files(self, files, reset=True):
        if reset:
            # This is necessary to prevent an error when closing Spyder with
            # mixed local and remote consoles.
            try:
                self.model.setRowCount(0)
            except RuntimeError:
                pass

        if files:
            logger.debug(f"Setting {len(files)} files")
            root = self.model.invisibleRootItem()

            more_files_items = self.model.match(
                self.model.index(0, 0), Qt.DisplayRole, _("Show more files")
            )
            if len(more_files_items):
                # Remove more files item
                self.model.removeRow(more_files_items[-1].row())

            more_files_available = self.model.match(
                self.model.index(0, 0),
                Qt.DisplayRole,
                _("Maximum number of files to display reached!"),
            )
            if len(more_files_available):
                # Remove more items available item
                self.model.removeRow(more_files_available[-1].row())

            for file in files:
                path = file["name"]
                name = os.path.relpath(path, self.root_prefix[self.server_id])

                file_type = file["type"]
                icon = ima.icon("FileIcon")
                if file_type == "directory":
                    icon = ima.icon("DirClosedIcon")

                file_name = QStandardItem(icon, name)
                file_name.setData(file)
                file_name.setToolTip(file["name"])

                if file_type == "directory":
                    file_size = QStandardItem("")
                else:
                    file_size = QStandardItem(
                        self._format_file_size(file["size"])
                    )
                file_type = QStandardItem(file_type)
                file_date_modified = QStandardItem(
                    datetime.fromtimestamp(file["mtime"]).strftime(
                        "%d/%m/%Y %I:%M %p"
                    )
                )

                items = [
                    file_name,
                    file_size,
                    file_type,
                    file_date_modified,
                ]
                for standard_item in items:
                    standard_item.setEditable(False)
                root.appendRow(items)

            # Add fetch more or more items available item
            if len(self.extra_files):
                fetch_more_item = QStandardItem(_("Show more files"))
                fetch_more_item.setEditable(False)
                fetch_more_item.setData(
                    {"name": "FETCH_MORE", "type": "ACTION"}
                )
                root.appendRow(fetch_more_item)
                self.view.setFirstColumnSpanned(
                    fetch_more_item.index().row(), root.index(), True
                )
            elif len(self.extra_files) == 0 and self.more_files_available:
                more_items_available = QStandardItem(
                    _("Maximum number of files to display reached!")
                )
                more_items_available.setEditable(False)
                more_items_available.setData(
                    {"name": "MESSAGE", "type": "ACTION"}
                )
                root.appendRow(more_items_available)
                self.view.setFirstColumnSpanned(
                    more_items_available.index().row(), root.index(), True
                )

            self.view.resizeColumnToContents(0)

    def fetch_more_files(self):
        fetch_files_display = self.get_conf("fetch_files_display")
        new_files = self.extra_files[:fetch_files_display]
        del self.extra_files[:fetch_files_display]
        self.set_files(new_files, reset=False)
        logger.debug(
            f"{len(self.extra_files)} extra files remaining to be shown"
        )

    def set_current_folder(self, folder):
        self.root_prefix[self.server_id] = folder
        return self.model.invisibleRootItem()

    def get_current_folder(self):
        return self.root_prefix[self.server_id]

    def go_to_parent_directory(self):
        parent_directory = os.path.dirname(self.root_prefix[self.server_id])
        logger.debug(
            f"Going to parent directory of {self.root_prefix[self.server_id]}: "
            f"{parent_directory}"
        )
        self.chdir(parent_directory)

    def go_to_previous_directory(self):
        self.histindex -= 1
        logger.debug(
            f"Going to previous directory in history with index "
            f"{self.histindex}"
        )
        self.chdir(browsing_history=True)

    def go_to_next_directory(self):
        self.histindex += 1
        logger.debug(
            f"Going to next directory in history with index {self.histindex}"
        )
        self.chdir(browsing_history=True)

    def refresh(self, new_path=None, force_current=False):
        if force_current:
            if new_path is None:
                new_path = self.root_prefix.get(self.server_id)
            self._do_remote_ls(new_path, self.server_id).connect(
                self._on_remote_ls
            )

        self.previous_action.setEnabled(False)
        self.next_action.setEnabled(False)

        if self.histindex is not None:
            self.previous_action.setEnabled(self.histindex > 0)
            self.next_action.setEnabled(self.histindex < len(self.history) - 1)

    def set_single_click_to_open(self, value):
        if value:
            try:
                self.view.doubleClicked.disconnect(self._on_clicked_item)
            except TypeError:
                pass
            self.view.clicked.connect(self._on_clicked_item)
        else:
            try:
                self.view.clicked.disconnect(self._on_clicked_item)
            except TypeError:
                pass
            self.view.doubleClicked.connect(self._on_clicked_item)

    def filter_files(self, name_filters=None):
        """Filter files given the defined list of filters."""
        if name_filters is None:
            name_filters = self.get_conf("name_filters")
        self.name_filters = []
        if self.filter_on:
            self.name_filters = name_filters
        self.refresh(force_current=True)

    def change_filter_state(self):
        self.filter_on = not self.filter_on
        self.filter_button.setChecked(self.filter_on)
        self.filter_button.setToolTip(_("Filter filenames"))
        self.filter_files()

    def new_package(self):
        new_name, valid = QInputDialog.getText(
            self, _("New Python Package"), _("Name as:"), QLineEdit.Normal, ""
        )
        if valid:
            self._new_item(new_name, with_content=True)

    def new_module(self):
        new_name, valid = QInputDialog.getText(
            self, _("New Python File"), _("Name as:"), QLineEdit.Normal, ".py"
        )
        if valid:
            self._new_item(new_name, for_file=True, with_content=True)

    def new_directory(self):
        new_name, valid = QInputDialog.getText(
            self, _("New Folder"), _("Name as:"), QLineEdit.Normal, ""
        )
        if valid:
            self._new_item(new_name)

    def new_file(self):
        new_name, valid = QInputDialog.getText(
            self, _("New File"), _("Name as:"), QLineEdit.Normal, ""
        )
        if valid:
            self._new_item(new_name, for_file=True)

    def copy_items(self):
        indexes = self._get_selected_indexes()
        if not indexes:
            return

        self._files_to_copy[self.server_id] = []
        for index in indexes:
            source_index = self.proxy_model.mapToSource(index)
            data_index = self.model.index(source_index.row(), 0)
            data = self.model.data(data_index, Qt.UserRole + 1)
            if data:
                self._files_to_copy[self.server_id].append(data["name"])

    def paste_items(self):
        if not self._files_to_copy.get(self.server_id, []):
            return

        if self.server_id not in self._files_to_paste:
            self._files_to_paste[self.server_id] = 0
        self._files_to_paste[self.server_id] += len(self._files_to_copy)
        self.sig_start_spinner_requested.emit()

        check_for_files_existence = True
        if self._files_to_paste[self.server_id] > 50:
            answer = QMessageBox.warning(
                self,
                _("Existence check"),
                _(
                    "You are about to paste more than 50 files in server "
                    "<b>{}</b>."
                    "<br><br>"
                    "Do you prefer to skip checking whether they exist in the "
                    "server to paste them more quickly?"
                ).format(self._get_server_name()),
                QMessageBox.Yes | QMessageBox.No,
            )

            if answer == QMessageBox.Yes:
                check_for_files_existence = False

        if check_for_files_existence:
            # Check if the remote files exist first. If some of them do, then
            # ask the user if they want to overwrite them with the files to be
            # pasted.
            self._check_if_remote_files_exist(
                self._files_to_copy[self.server_id]
            ).connect(
                AsyncDispatcher.QtSlot(
                    functools.partial(
                        self._on_check_if_remote_files_exist,
                        operation=RemoteExistenceOperations.Paste,
                    )
                )
            )
        else:
            for file in self._files_to_copy[self.server_id]:
                new_path = posixpath.join(
                    self.root_prefix[self.server_id], os.path.basename(file)
                )
                self._do_remote_paste(file, new_path).connect(
                    self._on_remote_paste
                )

    def rename_items(self):
        indexes = self._get_selected_indexes()
        if not indexes:
            return

        if self.server_id not in self._files_to_rename:
            self._files_to_rename[self.server_id] = 0
        self._files_to_rename[self.server_id] += len(indexes)
        self.sig_start_spinner_requested.emit()

        for index in indexes:
            source_index = self.proxy_model.mapToSource(index)
            data_index = self.model.index(source_index.row(), 0)
            data = self.model.data(data_index, Qt.UserRole + 1)
            if data:
                old_path = data["name"]
                relpath = os.path.relpath(
                    old_path, self.root_prefix[self.server_id]
                )
                new_relpath, valid = QInputDialog.getText(
                    self,
                    _("Rename"),
                    _("New name for <b>{}</b>:").format(relpath),
                    QLineEdit.Normal,
                    relpath,
                )
                if valid:
                    new_path = posixpath.join(
                        self.root_prefix[self.server_id], new_relpath
                    )
                    self._do_remote_rename(old_path, str(new_path)).connect(
                        self._on_remote_rename
                    )

    def copy_paths(self):
        paths = []
        indexes = self._get_selected_indexes()
        if not indexes or not self.view.currentIndex().isValid():
            paths.append(self.root_prefix[self.server_id])
        else:
            for index in indexes:
                source_index = self.proxy_model.mapToSource(index)
                data_index = self.model.index(source_index.row(), 0)
                data = self.model.data(data_index, Qt.UserRole + 1)
                if data:
                    paths.append(data["name"])

        if len(paths) > 1:
            # TODO: Decide what format we want for multiple paths (i.e. local
            # explorer or this one).
            clipboard_paths = ' '.join(['"' + path + '"' for path in paths])
        else:
            clipboard_paths = '"' + paths[0] + '"'

        cb = QApplication.clipboard()
        cb.setText(clipboard_paths, mode=QClipboard.Mode.Clipboard)

    def delete_items(self):
        yes_to_all = False
        indexes = self._get_selected_indexes()
        if not indexes:
            return

        if len(indexes) > 1:
            buttons = (
                QMessageBox.Yes
                | QMessageBox.YesToAll
                | QMessageBox.No
                | QMessageBox.Cancel
            )
        else:
            buttons = QMessageBox.Yes | QMessageBox.No

        if self.server_id not in self._files_to_delete:
            self._files_to_delete[self.server_id] = 0
        self._files_to_delete[self.server_id] += len(indexes)
        self.sig_start_spinner_requested.emit()

        for index in indexes:
            source_index = self.proxy_model.mapToSource(index)
            data_index = self.model.index(source_index.row(), 0)
            data = self.model.data(data_index, Qt.UserRole + 1)
            if data:
                path = data["name"]
                filename = os.path.relpath(
                    path, self.root_prefix[self.server_id]
                )

                if not yes_to_all:
                    result = QMessageBox.warning(
                        self,
                        _("Delete"),
                        _(
                            "Do you really want to delete <b>{filename}</b>?"
                            "<br><br>"
                            "<b>Note</b>: The file/directory will be removed "
                            "permanently from the remote filesystem."
                        ).format(filename=filename),
                        buttons
                    )

                if result == QMessageBox.YesToAll:
                    yes_to_all = True
                elif result == QMessageBox.No:
                    if self._files_to_delete[self.server_id] > 0:
                        self._files_to_delete[self.server_id] -= 1

                    if not self._operation_in_progress:
                        self.sig_stop_spinner_requested.emit()

                    continue
                elif result == QMessageBox.Cancel:
                    self._files_to_delete[self.server_id] = 0

                    if not self._operation_in_progress:
                        self.sig_stop_spinner_requested.emit()

                    return

                if result == QMessageBox.Yes or yes_to_all:
                    is_file = data["type"] == "file"
                    self._do_remote_delete(path, is_file=is_file).connect(
                        self._on_remote_delete
                    )

    def download_items(self):
        indexes = self._get_selected_indexes()
        if not indexes:
            return

        # Directory where the files will be downloaded
        directory = getexistingdirectory(
            self, _("Download directory"), getcwd_or_home()
        )

        # This is necessary in case the user clicks Cancel in the dialog
        if not directory:
            return

        # Keep track of how many files will be downloaded to stop the spinner
        # when all of them are processed.
        if self.server_id not in self._files_to_download:
            self._files_to_download[self.server_id] = 0
        self._files_to_download[self.server_id] += len(indexes)
        self.sig_start_spinner_requested.emit()

        # Check if users want to overwrite all downloaded files
        yes_to_all = False

        # Buttons to show in the dialog that asks to overwrite files
        if len(indexes) > 1:
            buttons = (
                QMessageBox.Yes
                | QMessageBox.YesToAll
                | QMessageBox.No
                | QMessageBox.Cancel
            )
        else:
            buttons = QMessageBox.Yes | QMessageBox.No

        for index in indexes:
            source_index = self.proxy_model.mapToSource(index)
            data_index = self.model.index(source_index.row(), 0)
            data = self.model.data(data_index, Qt.UserRole + 1)
            if data:
                path = data["name"]
                filename = os.path.relpath(
                    path, self.root_prefix[self.server_id]
                )
                is_file = data["type"] == "file"
                remote_filename = filename if is_file else f"{filename}.zip"

                # Local filename
                local_filename = os.path.join(directory, remote_filename)

                # Check if a local file with the same name exists
                if not yes_to_all and os.path.exists(local_filename):
                    answer = QMessageBox.warning(
                        self,
                        _("Overwrite file"),
                        _(
                            "The file <b>{}</b> that you're trying to "
                            "download from server <b>{}</b> already exists in "
                            "the location you selected."
                            "<br><br>"
                            "Do you want to overwrite it?"
                        ).format(remote_filename, self._get_server_name()),
                        buttons,
                    )

                    if answer == QMessageBox.YesToAll:
                        yes_to_all = True
                    elif answer == QMessageBox.No:
                        if self._files_to_download[self.server_id] > 0:
                            self._files_to_download[self.server_id] -= 1

                        if not self._operation_in_progress:
                            self.sig_stop_spinner_requested.emit()

                        continue
                    elif answer == QMessageBox.Cancel:
                        self._files_to_download[self.server_id] = 0

                        if not self._operation_in_progress:
                            self.sig_stop_spinner_requested.emit()

                        return

                # Download file or directory
                method = (
                    self._do_remote_download_file
                    if is_file
                    else self._do_remote_download_directory
                )
                method(path).connect(
                    AsyncDispatcher.QtSlot(
                        functools.partial(
                            self._on_remote_download_file,
                            remote_filename=remote_filename,
                            local_directory=directory,
                            is_file=is_file,
                        )
                    )
                )

    def upload_files(self):
        local_paths, __ = getopenfilenames(
            self,
            _("Upload files"),
            getcwd_or_home(),
            _("All files") + " (*)",
        )

        local_paths = [path for path in local_paths if os.path.exists(path)]
        if not local_paths:
            return

        if self.server_id not in self._files_to_upload:
            self._files_to_upload[self.server_id] = 0
        self._files_to_upload[self.server_id] += len(local_paths)
        self.sig_start_spinner_requested.emit()

        check_for_files_existence = True
        if len(local_paths) > 50:
            answer = QMessageBox.warning(
                self,
                _("Existence check"),
                _(
                    "You are about to upload more than 50 files to server "
                    "<b>{}</b>."
                    "<br><br>"
                    "Do you prefer to skip checking whether they exist on the "
                    "server to upload them more quickly?"
                ).format(self._get_server_name()),
                QMessageBox.Yes | QMessageBox.No,
            )

            if answer == QMessageBox.Yes:
                check_for_files_existence = False

        if check_for_files_existence:
            # Check if the remote files exist first. If some of them do, then
            # ask the user if they want to overwrite them with the local files
            # to be uploaded.
            self._check_if_remote_files_exist(local_paths).connect(
                AsyncDispatcher.QtSlot(
                    functools.partial(
                        self._on_check_if_remote_files_exist,
                        operation=RemoteExistenceOperations.Upload,
                    )
                )
            )
        else:
            for path in local_paths:
                self._do_remote_upload_file(path).connect(
                    self._on_remote_upload_file
                )

    def reset(self, server_id):
        self.root_prefix[server_id] = None
